from datetime import datetime

import mongoengine
import mongomock
from mongomock import gridfs

gridfs.enable_gridfs_integration()
mongoengine.connect(
    "graphene-mongo-test", mongo_client_class=mongomock.MongoClient, alias="default"
)


# mongoengine.connect('graphene-mongo-test', host='mongodb://localhost/graphene-mongo-dev')


class Publisher(mongoengine.Document):
    meta = {"collection": "test_publisher"}
    name = mongoengine.StringField()

    @property
    def legal_name(self):
        return self.name + " Inc."

    def bad_field(self):
        return None


class Editor(mongoengine.Document):
    """
    An Editor of a publication.
    """

    meta = {"collection": "test_editor"}
    id = mongoengine.StringField(primary_key=True)
    first_name = mongoengine.StringField(
        required=True, help_text="Editor's first name.", db_field="fname"
    )
    last_name = mongoengine.StringField(required=True, help_text="Editor's last name.")
    metadata = mongoengine.MapField(
        field=mongoengine.StringField(), help_text="Arbitrary metadata."
    )
    company = mongoengine.LazyReferenceField(Publisher)
    avatar = mongoengine.FileField()
    seq = mongoengine.SequenceField()


class Article(mongoengine.Document):
    meta = {"collection": "test_article"}
    headline = mongoengine.StringField(required=True, help_text="The article headline.")
    pub_date = mongoengine.DateTimeField(
        default=datetime.now,
        verbose_name="publication date",
        help_text="The date of first press.",
    )
    editor = mongoengine.ReferenceField(Editor)
    reporter = mongoengine.ReferenceField("Reporter")
    # Will not convert these fields cause no choices
    # generic_reference = mongoengine.GenericReferenceField()
    # generic_embedded_document = mongoengine.GenericEmbeddedDocumentField()


class EmbeddedArticle(mongoengine.EmbeddedDocument):
    meta = {"collection": "test_embedded_article"}
    headline = mongoengine.StringField(required=True)
    pub_date = mongoengine.DateTimeField(default=datetime.now)
    editor = mongoengine.ReferenceField(Editor)
    reporter = mongoengine.ReferenceField("Reporter")


class EmbeddedFoo(mongoengine.EmbeddedDocument):
    meta = {"collection": "test_embedded_foo"}
    bar = mongoengine.StringField()


class Reporter(mongoengine.Document):
    meta = {"collection": "test_reporter"}
    id = mongoengine.StringField(primary_key=True)
    first_name = mongoengine.StringField(required=True)
    last_name = mongoengine.StringField(required=True)
    email = mongoengine.EmailField()
    awards = mongoengine.ListField(mongoengine.StringField())
    articles = mongoengine.ListField(mongoengine.ReferenceField(Article))
    embedded_articles = mongoengine.ListField(
        mongoengine.EmbeddedDocumentField(EmbeddedArticle),
    )
    embedded_list_articles = mongoengine.EmbeddedDocumentListField(EmbeddedArticle)
    generic_reference = mongoengine.GenericReferenceField(choices=[Article, Editor], required=True)
    generic_embedded_document = mongoengine.GenericEmbeddedDocumentField(
        choices=[EmbeddedArticle, EmbeddedFoo]
    )
    generic_references = mongoengine.ListField(
        mongoengine.GenericReferenceField(choices=[Article, Editor])
    )


class Player(mongoengine.Document):
    meta = {"collection": "test_player"}
    first_name = mongoengine.StringField(required=True)
    last_name = mongoengine.StringField(required=True)
    opponent = mongoengine.ReferenceField("Player")
    players = mongoengine.ListField(mongoengine.ReferenceField("Player"))
    articles = mongoengine.ListField(mongoengine.ReferenceField("Article"))
    embedded_list_articles = mongoengine.EmbeddedDocumentListField(EmbeddedArticle)


class Parent(mongoengine.Document):
    meta = {"collection": "test_parent", "allow_inheritance": True}
    bar = mongoengine.StringField()
    loc = mongoengine.MultiPolygonField()


class CellTower(mongoengine.Document):
    meta = {"collection": "test_cell_tower"}
    code = mongoengine.StringField()
    base = mongoengine.PolygonField()
    coverage_area = mongoengine.MultiPolygonField()


class Child(Parent):
    meta = {"collection": "test_parent"}
    baz = mongoengine.StringField()
    loc = mongoengine.PointField()


class AnotherChild(Parent):
    meta = {"collection": "test_parent"}
    qux = mongoengine.StringField()
    loc = mongoengine.PointField()


class ProfessorMetadata(mongoengine.EmbeddedDocument):
    meta = {"collection": "test_professor_metadata"}
    id = mongoengine.StringField(primary_key=False)
    first_name = mongoengine.StringField()
    last_name = mongoengine.StringField()
    departments = mongoengine.ListField(mongoengine.StringField())


class ProfessorVector(mongoengine.Document):
    meta = {"collection": "test_professor_vector"}
    vec = mongoengine.ListField(mongoengine.FloatField())
    metadata = mongoengine.EmbeddedDocumentField(ProfessorMetadata)


class ParentWithRelationship(mongoengine.Document):
    meta = {"collection": "test_parent_reference"}
    before_child = mongoengine.ListField(
        mongoengine.ReferenceField("ChildRegisteredBefore"),
    )
    after_child = mongoengine.ListField(
        mongoengine.ReferenceField("ChildRegisteredAfter"),
    )
    name = mongoengine.StringField()


class ChildRegisteredBefore(mongoengine.Document):
    meta = {"collection": "test_child_before_reference"}
    parent = mongoengine.ReferenceField(ParentWithRelationship)
    name = mongoengine.StringField()


class ChildRegisteredAfter(mongoengine.Document):
    meta = {"collection": "test_child_after_reference"}
    parent = mongoengine.ReferenceField(ParentWithRelationship)
    name = mongoengine.StringField()


class ErroneousModel(mongoengine.Document):
    meta = {"collection": "test_colliding_objects_model"}

    objects = mongoengine.ListField(mongoengine.StringField())


class Bar(mongoengine.EmbeddedDocument):
    some_list_field = mongoengine.ListField(mongoengine.StringField(), required=True)


class Foo(mongoengine.Document):
    bars = mongoengine.EmbeddedDocumentListField(Bar)



from graphene import relay
from graphene_mongo import MongoengineObjectType
from .models import Bike, Shop


class BikeType(MongoengineObjectType):
    class Meta:
        model = Bike
        interfaces = (relay.Node,)


class ShopType(MongoengineObjectType):
    class Meta:
        model = Shop



import graphene
from graphene.relay import Node

from . import models
from . import types  # noqa: F401
from .models import ProfessorMetadata
from ..types_async import AsyncMongoengineObjectType


class PublisherAsyncNode(AsyncMongoengineObjectType):
    legal_name = graphene.String()
    bad_field = graphene.String()

    class Meta:
        model = models.Publisher
        only_fields = ("id", "name")
        interfaces = (Node,)


class ArticleAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.Article
        interfaces = (Node,)


class EditorAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.Editor
        interfaces = (Node,)


class EmbeddedArticleAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.EmbeddedArticle
        interfaces = (Node,)


class PlayerAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.Player
        interfaces = (Node,)
        filter_fields = {"first_name": ["istartswith", "in"]}


class ReporterAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.Reporter
        interfaces = (Node,)


class ParentAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.Parent
        interfaces = (Node,)


class ChildAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.Child
        interfaces = (Node,)


class ChildRegisteredBeforeAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.ChildRegisteredBefore
        interfaces = (Node,)


class ChildRegisteredAfterAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.ChildRegisteredAfter
        interfaces = (Node,)


class ParentWithRelationshipAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.ParentWithRelationship
        interfaces = (Node,)


class ProfessorMetadataAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = ProfessorMetadata
        interfaces = (graphene.Node,)


class ProfessorVectorAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.ProfessorVector
        interfaces = (Node,)


class ErroneousModelAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.ErroneousModel
        interfaces = (Node,)


class BarAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.Bar
        interfaces = (Node,)


class FooAsyncNode(AsyncMongoengineObjectType):
    class Meta:
        model = models.Foo
        interfaces = (Node,)



from pytest import raises

from graphene import Field, Int, Interface, ObjectType
from graphene.relay import Node, is_node

from .. import registry
from ..types import MongoengineObjectType, MongoengineObjectTypeOptions
from .models import Article, EmbeddedArticle, Reporter
from .models import Parent, Child
from .utils import with_local_registry

registry.reset_global_registry()


class Human(MongoengineObjectType):
    pub_date = Int()

    class Meta:
        model = Article
        registry = registry.get_global_registry()
        interfaces = (Node,)


class Being(MongoengineObjectType):
    class Meta:
        model = EmbeddedArticle
        interfaces = (Node,)


class Character(MongoengineObjectType):
    class Meta:
        model = Reporter
        registry = registry.get_global_registry()


class Dad(MongoengineObjectType):
    class Meta:
        model = Parent
        registry = registry.get_global_registry()


class Son(MongoengineObjectType):
    class Meta:
        model = Child
        registry = registry.get_global_registry()


def test_mongoengine_interface():
    assert issubclass(Node, Interface)
    assert issubclass(Node, Node)


def test_objecttype_registered():
    assert issubclass(Character, ObjectType)
    assert Character._meta.model == Reporter
    assert set(Character._meta.fields.keys()) == set(
        [
            "id",
            "first_name",
            "last_name",
            "email",
            "embedded_articles",
            "embedded_list_articles",
            "articles",
            "awards",
            "generic_reference",
            "generic_embedded_document",
            "generic_references",
        ]
    )


def test_mongoengine_inheritance():
    assert issubclass(Son._meta.model, Dad._meta.model)


def test_node_replacedfield():
    idfield = Human._meta.fields["pub_date"]
    assert isinstance(idfield, Field)
    assert idfield.type == Int


def test_object_type():
    assert issubclass(Human, ObjectType)
    assert set(Human._meta.fields.keys()) == set(
        ["id", "headline", "pub_date", "editor", "reporter"]
    )
    assert is_node(Human)


def test_should_raise_if_no_model():
    with raises(Exception) as excinfo:

        class Human1(MongoengineObjectType):
            pass

    assert "valid Mongoengine Model" in str(excinfo.value)


def test_should_raise_if_model_is_invalid():
    with raises(Exception) as excinfo:

        class Human2(MongoengineObjectType):
            class Meta:
                model = 1

    assert "valid Mongoengine Model" in str(excinfo.value)


@with_local_registry
def test_mongoengine_objecttype_only_fields():
    class A(MongoengineObjectType):
        class Meta:
            model = Article
            only_fields = "headline"

    fields = set(A._meta.fields.keys())
    assert fields == set(["headline"])


@with_local_registry
def test_mongoengine_objecttype_exclude_fields():
    class A(MongoengineObjectType):
        class Meta:
            model = Article
            exclude_fields = "headline"

    assert "headline" not in list(A._meta.fields.keys())


@with_local_registry
def test_mongoengine_objecttype_order_by():
    class A(MongoengineObjectType):
        class Meta:
            model = Article
            order_by = "some_order_by_statement"

    assert "some_order_by_statement" not in list(A._meta.fields.keys())


@with_local_registry
def test_passing_meta_when_subclassing_mongoengine_objecttype():
    class TypeSubclassWithBadOptions(MongoengineObjectType):
        class Meta:
            abstract = True

        @classmethod
        def __init_subclass_with_meta__(cls, **kwargs):
            _meta = ["hi"]
            super(TypeSubclassWithBadOptions, cls).__init_subclass_with_meta__(
                _meta=_meta, **kwargs
            )

    with raises(Exception) as einfo:

        class A(TypeSubclassWithBadOptions):
            class Meta:
                model = Article

    assert "instance of MongoengineGenericObjectTypeOptions" in str(einfo.value)

    class TypeSubclass(MongoengineObjectType):
        class Meta:
            abstract = True

        @classmethod
        def __init_subclass_with_meta__(cls, some_subclass_attr=None, **kwargs):
            _meta = MongoengineObjectTypeOptions(cls)
            _meta.some_subclass_attr = some_subclass_attr
            super(TypeSubclass, cls).__init_subclass_with_meta__(_meta=_meta, **kwargs)

    class B(TypeSubclass):
        class Meta:
            model = Article
            some_subclass_attr = "someval"

    assert hasattr(B._meta, "some_subclass_attr")
    assert B._meta.some_subclass_attr == "someval"



import graphene
from django.core.exceptions import ObjectDoesNotExist
from .models import Bike
from .types import BikeType


class BikeInput(graphene.InputObjectType):
    id = graphene.ID()
    name = graphene.String()
    brand = graphene.String()
    year = graphene.String()
    size = graphene.List(graphene.String)
    wheel_size = graphene.Float()
    type = graphene.String()


class CreateBikeMutation(graphene.Mutation):
    bike = graphene.Field(BikeType)

    class Arguments:
        bike_data = BikeInput(required=True)

    def mutate(self, info, bike_data=None):
        bike = Bike(
            name=bike_data.name,
            brand=bike_data.brand,
            year=bike_data.year,
            size=bike_data.size,
            wheel_size=bike_data.wheel_size,
            type=bike_data.type,
        )
        bike.save()

        return CreateBikeMutation(bike=bike)


class UpdateBikeMutation(graphene.Mutation):
    bike = graphene.Field(BikeType)

    class Arguments:
        bike_data = BikeInput(required=True)

    @staticmethod
    def get_object(id):
        return Bike.objects.get(pk=id)

    def mutate(self, info, bike_data=None):
        bike = UpdateBikeMutation.get_object(bike_data.id)
        if bike_data.name:
            bike.name = bike_data.name
        if bike_data.brand:
            bike.brand = bike_data.brand
        if bike_data.year:
            bike.year = bike_data.year
        if bike_data.size:
            bike.size = bike_data.size
        if bike_data.wheel_size:
            bike.wheel_size = bike_data.wheel_size
        if bike_data.type:
            bike.type = bike_data.type

        bike.save()

        return UpdateBikeMutation(bike=bike)


class DeleteBikeMutation(graphene.Mutation):
    class Arguments:
        id = graphene.ID(required=True)

    success = graphene.Boolean()

    def mutate(self, info, id):
        try:
            Bike.objects.get(pk=id).delete()
            success = True
        except ObjectDoesNotExist:
            success = False

        return DeleteBikeMutation(success=success)



import base64
import os
import json
import graphene
import pytest

from . import models
from . import types


@pytest.mark.asyncio
async def test_should_query_editor(fixtures, fixtures_dirname):
    class Query(graphene.ObjectType):
        editor = graphene.Field(types.EditorType)
        editors = graphene.List(types.EditorType)

        async def resolve_editor(self, *args, **kwargs):
            return models.Editor.objects.first()

        async def resolve_editors(self, *args, **kwargs):
            return list(models.Editor.objects.all())

    query = """
        query EditorQuery {
            editor {
                firstName,
                metadata,
                company {
                    name
                },
                avatar {
                    contentType,
                    chunkSize,
                    length,
                    data
                }
            }
            editors {
                firstName,
                lastName
            }
        }
    """

    avator_filename = os.path.join(fixtures_dirname, "image.jpg")
    with open(avator_filename, "rb") as f:
        data = base64.b64encode(f.read())

    expected = {
        "editor": {
            "firstName": "Penny",
            "company": {"name": "Newsco"},
            "avatar": {
                "contentType": "image/jpeg",
                "chunkSize": 261120,
                "length": 46928,
                "data": data.decode("utf-8"),
            },
        },
        "editors": [
            {"firstName": "Penny", "lastName": "Hardaway"},
            {"firstName": "Grant", "lastName": "Hill"},
            {"firstName": "Dennis", "lastName": "Rodman"},
        ],
    }
    expected_metadata = {"age": "20", "nickname": "$1"}

    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(query)
    assert not result.errors
    metadata = result.data["editor"].pop("metadata")
    assert json.loads(metadata) == expected_metadata
    assert result.data == expected


@pytest.mark.asyncio
async def test_should_query_reporter(fixtures):
    class Query(graphene.ObjectType):
        reporter = graphene.Field(types.ReporterType)

        async def resolve_reporter(self, *args, **kwargs):
            return models.Reporter.objects.first()

    query = """
        query ReporterQuery {
            reporter {
                firstName,
                lastName,
                email,
                articles {
                    headline
                },
                embeddedArticles {
                    headline
                },
                embeddedListArticles {
                    headline
                },
                awards
            }
        }
    """
    expected = {
        "reporter": {
            "firstName": "Allen",
            "lastName": "Iverson",
            "email": "ai@gmail.com",
            "articles": [{"headline": "Hello"}, {"headline": "World"}],
            "embeddedArticles": [{"headline": "Real"}, {"headline": "World"}],
            "embeddedListArticles": [{"headline": "World"}, {"headline": "Real"}],
            "awards": ["2010-mvp"],
        }
    }

    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(query)
    assert not result.errors
    assert result.data == expected


@pytest.mark.asyncio
async def test_should_custom_kwargs(fixtures):
    class Query(graphene.ObjectType):
        editors = graphene.List(types.EditorType, first=graphene.Int())

        async def resolve_editors(self, *args, **kwargs):
            editors = models.Editor.objects()
            if "first" in kwargs:
                editors = editors[: kwargs["first"]]
            return list(editors)

    query = """
        query EditorQuery {
            editors(first: 2) {
                firstName,
                lastName
            }
        }
    """
    expected = {
        "editors": [
            {"firstName": "Penny", "lastName": "Hardaway"},
            {"firstName": "Grant", "lastName": "Hill"},
        ]
    }
    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(query)
    assert not result.errors
    assert result.data == expected


@pytest.mark.asyncio
async def test_should_self_reference(fixtures):
    class Query(graphene.ObjectType):
        all_players = graphene.List(types.PlayerType)

        async def resolve_all_players(self, *args, **kwargs):
            return models.Player.objects.all()

    query = """
        query PlayersQuery {
            allPlayers {
                firstName,
                opponent {
                    firstName
                },
                players {
                    firstName
                }
            }
        }
    """
    expected = {
        "allPlayers": [
            {
                "firstName": "Michael",
                "opponent": None,
                "players": [{"firstName": "Magic"}],
            },
            {
                "firstName": "Magic",
                "opponent": {"firstName": "Michael"},
                "players": [{"firstName": "Michael"}],
            },
            {
                "firstName": "Larry",
                "opponent": None,
                "players": [{"firstName": "Michael"}, {"firstName": "Magic"}],
            },
            {"firstName": "Chris", "opponent": None, "players": []},
        ]
    }
    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(query)
    assert not result.errors
    assert result.data == expected


@pytest.mark.asyncio
async def test_should_query_with_embedded_document(fixtures):
    class Query(graphene.ObjectType):
        professor_vector = graphene.Field(types.ProfessorVectorType, id=graphene.String())

        async def resolve_professor_vector(self, info, id):
            return models.ProfessorVector.objects(metadata__id=id).first()

    query = """
        query {
          professorVector(id: "5e06aa20-6805-4eef-a144-5615dedbe32b") {
            vec
            metadata {
                firstName
            }
          }
        }
    """

    expected = {"professorVector": {"vec": [1.0, 2.3], "metadata": {"firstName": "Steven"}}}
    schema = graphene.Schema(query=Query, types=[types.ProfessorVectorType])
    result = await schema.execute_async(query)
    assert not result.errors
    assert result.data == expected


@pytest.mark.asyncio
async def test_should_query_child(fixtures):
    class Query(graphene.ObjectType):
        children = graphene.List(types.ChildType)

        async def resolve_children(self, *args, **kwargs):
            return list(models.Child.objects.all())

    query = """
        query Query {
            children {
                bar,
                baz,
                loc {
                     type,
                     coordinates
                }
            }
        }
    """
    expected = {
        "children": [
            {"bar": "BAR", "baz": "BAZ", "loc": None},
            {
                "bar": "bar",
                "baz": "baz",
                "loc": {"type": "Point", "coordinates": [10.0, 20.0]},
            },
        ]
    }

    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(query)
    assert not result.errors
    assert result.data == expected


@pytest.mark.asyncio
async def test_should_query_other_childs(fixtures):
    class Query(graphene.ObjectType):
        children = graphene.List(types.AnotherChildType)

        async def resolve_children(self, *args, **kwargs):
            return list(models.AnotherChild.objects.all())

    query = """
        query Query {
            children {
                bar,
                qux,
                loc {
                     type,
                     coordinates
                }
            }
        }
    """
    expected = {
        "children": [
            {"bar": "BAR", "qux": "QUX", "loc": None},
            {
                "bar": "bar",
                "qux": "qux",
                "loc": {"type": "Point", "coordinates": [20, 10]},
            },
        ]
    }

    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(query)
    assert not result.errors
    assert result.data == expected


@pytest.mark.asyncio
async def test_should_query_all_childs(fixtures):
    class Query(graphene.ObjectType):
        children = graphene.List(types.ChildUnionType)

        async def resolve_children(self, *args, **kwargs):
            return list(models.Parent.objects.all())

    query = """
        query Query {
            children {
                ... on ParentInterface {
                    bar
                }
                ... on ChildType{
                    baz
                    loc {
                        type,
                        coordinates
                    }
                }
                ... on AnotherChildType {
                    qux
                    loc {
                        type,
                        coordinates
                    }
                }
            }
        }
    """
    expected = {
        "children": [
            {"bar": "BAR", "baz": "BAZ", "loc": None},
            {
                "bar": "bar",
                "baz": "baz",
                "loc": {"type": "Point", "coordinates": [10.0, 20.0]},
            },
            {"bar": "BAR", "qux": "QUX", "loc": None},
            {
                "bar": "bar",
                "qux": "qux",
                "loc": {"type": "Point", "coordinates": [20, 10]},
            },
        ]
    }

    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(query)
    assert not result.errors
    assert result.data == expected


@pytest.mark.asyncio
async def test_should_query_cell_tower(fixtures):
    class Query(graphene.ObjectType):
        cell_towers = graphene.List(types.CellTowerType)

        async def resolve_cell_towers(self, *args, **kwargs):
            return list(models.CellTower.objects.all())

    query = """
        query Query {
            cellTowers {
                code,
                base {
                    type,
                    coordinates
                },
                coverageArea {
                     type,
                     coordinates
                }
            }
        }
    """
    expected = {
        "cellTowers": [
            {
                "code": "bar",
                "base": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [-43.36556, -22.99669],
                            [-43.36539, -23.01928],
                            [-43.26583, -23.01802],
                            [-43.36717, -22.98855],
                            [-43.36636, -22.99351],
                            [-43.36556, -22.99669],
                        ]
                    ],
                },
                "coverageArea": {
                    "type": "MultiPolygon",
                    "coordinates": [
                        [
                            [
                                [-43.36556, -22.99669],
                                [-43.36539, -23.01928],
                                [-43.26583, -23.01802],
                                [-43.36717, -22.98855],
                                [-43.36636, -22.99351],
                                [-43.36556, -22.99669],
                            ]
                        ]
                    ],
                },
            }
        ]
    }

    schema = graphene.Schema(query=Query)
    result = await schema.execute_async(query)
    assert not result.errors
    assert result.data == expected



from __future__ import absolute_import

from functools import partial
from itertools import filterfalse
from typing import Coroutine

import bson
import graphene
import mongoengine
import pymongo
from bson import DBRef, ObjectId
from graphene import Context
from graphene.relay import ConnectionField
from graphene.utils.str_converters import to_snake_case
from graphql import GraphQLResolveInfo
from graphql_relay import cursor_to_offset, from_global_id
from mongoengine import QuerySet
from promise import Promise
from pymongo.errors import OperationFailure

from . import MongoengineConnectionField
from .registry import get_global_async_registry
from .utils import (
    ExecutorEnum,
    connection_from_iterables,
    find_skip_and_limit,
    get_query_fields,
    sync_to_async,
    has_page_info,
)

PYMONGO_VERSION = tuple(pymongo.version_tuple[:2])


class AsyncMongoengineConnectionField(MongoengineConnectionField):
    def __init__(self, type, *args, **kwargs):
        super(AsyncMongoengineConnectionField, self).__init__(type, *args, **kwargs)

    @property
    def executor(self):
        return ExecutorEnum.ASYNC

    @property
    def type(self):
        from .types_async import AsyncMongoengineObjectType

        _type = super(ConnectionField, self).type
        assert issubclass(
            _type, AsyncMongoengineObjectType
        ), "AsyncMongoengineConnectionField only accepts AsyncMongoengineObjectType types"
        assert _type._meta.connection, "The type {} doesn't have a connection".format(
            _type.__name__
        )
        return _type._meta.connection

    @property
    def fields(self):
        return super(AsyncMongoengineConnectionField, self).fields

    @property
    def registry(self):
        return getattr(self.node_type._meta, "registry", get_global_async_registry())

    async def default_resolver(self, _root, info, required_fields=None, resolved=None, **args):
        if required_fields is None:
            required_fields = list()
        args = args or {}
        for key, value in dict(args).items():
            if value is None:
                del args[key]
        if _root is not None and not resolved:
            field_name = to_snake_case(info.field_name)
            if not hasattr(_root, "_fields_ordered"):
                if isinstance(getattr(_root, field_name, []), list):
                    args["pk__in"] = [r.id for r in getattr(_root, field_name, [])]
            elif field_name in _root._fields_ordered and not (
                isinstance(_root._fields[field_name].field, mongoengine.EmbeddedDocumentField)
                or isinstance(
                    _root._fields[field_name].field,
                    mongoengine.GenericEmbeddedDocumentField,
                )
            ):
                if getattr(_root, field_name, []) is not None:
                    args["pk__in"] = [r.id for r in getattr(_root, field_name, [])]

        _id = args.pop("id", None)

        if _id is not None:
            args["pk"] = from_global_id(_id)[-1]
        iterables = []
        list_length = 0
        skip = 0
        count = 0
        limit = None
        reverse = False
        first = args.pop("first", None)
        after = args.pop("after", None)
        if after:
            after = cursor_to_offset(after)
        last = args.pop("last", None)
        before = args.pop("before", None)
        if before:
            before = cursor_to_offset(before)
        requires_page_info = has_page_info(info)
        has_next_page = False

        if resolved is not None:
            items = resolved

            if isinstance(items, QuerySet):
                try:
                    if last is not None and after is not None:
                        count = await sync_to_async(items.count)(with_limit_and_skip=False)
                    else:
                        count = None
                except OperationFailure:
                    count = await sync_to_async(len)(items)
            else:
                count = len(items)

            skip, limit, reverse = find_skip_and_limit(
                first=first, last=last, after=after, before=before, count=count
            )

            if isinstance(items, QuerySet):
                if limit:
                    _base_query: QuerySet = (
                        await sync_to_async(items.order_by("-pk").skip)(skip)
                        if reverse
                        else await sync_to_async(items.skip)(skip)
                    )
                    items = await sync_to_async(_base_query.limit)(limit)
                    has_next_page = (
                        (
                            await sync_to_async(len)(
                                await sync_to_async(_base_query.skip(limit).only("id").limit)(1)
                            )
                            != 0
                        )
                        if requires_page_info
                        else False
                    )
                elif skip:
                    items = await sync_to_async(items.skip)(skip)
            else:
                if limit:
                    if reverse:
                        _base_query = items[::-1]
                        items = _base_query[skip : skip + limit]
                    else:
                        _base_query = items
                        items = items[skip : skip + limit]
                    has_next_page = (
                        (skip + limit) < len(_base_query) if requires_page_info else False
                    )
                elif skip:
                    items = items[skip:]
            iterables = await sync_to_async(list)(items)
            list_length = len(iterables)

        elif callable(getattr(self.model, "objects", None)):
            if (
                _root is None
                or args
                or isinstance(getattr(_root, field_name, []), AsyncMongoengineConnectionField)
            ):
                args_copy = args.copy()
                for key in args.copy():
                    if key not in self.model._fields_ordered:
                        args_copy.pop(key)
                    elif (
                        isinstance(getattr(self.model, key), mongoengine.fields.ReferenceField)
                        or isinstance(
                            getattr(self.model, key),
                            mongoengine.fields.GenericReferenceField,
                        )
                        or isinstance(
                            getattr(self.model, key),
                            mongoengine.fields.LazyReferenceField,
                        )
                        or isinstance(
                            getattr(self.model, key),
                            mongoengine.fields.CachedReferenceField,
                        )
                    ):
                        if not isinstance(args_copy[key], ObjectId):
                            _from_global_id = from_global_id(args_copy[key])[1]
                            if bson.objectid.ObjectId.is_valid(_from_global_id):
                                args_copy[key] = ObjectId(_from_global_id)
                            else:
                                args_copy[key] = _from_global_id
                    elif isinstance(getattr(self.model, key), mongoengine.fields.EnumField):
                        if getattr(args_copy[key], "value", None):
                            args_copy[key] = args_copy[key].value

                if PYMONGO_VERSION >= (3, 7):
                    count = await sync_to_async(
                        (mongoengine.get_db()[self.model._get_collection_name()]).count_documents
                    )(args_copy)
                else:
                    count = await sync_to_async(self.model.objects(args_copy).count)()
                if count != 0:
                    skip, limit, reverse = find_skip_and_limit(
                        first=first, after=after, last=last, before=before, count=count
                    )
                    iterables = self.get_queryset(
                        self.model, info, required_fields, skip, limit, reverse, **args
                    )
                    iterables = await sync_to_async(list)(iterables)
                    list_length = len(iterables)
                    if isinstance(info, GraphQLResolveInfo):
                        if not info.context:
                            info = info._replace(context=Context())
                        info.context.queryset = self.get_queryset(
                            self.model, info, required_fields, **args
                        )

            elif "pk__in" in args and args["pk__in"]:
                count = len(args["pk__in"])
                skip, limit, reverse = find_skip_and_limit(
                    first=first, last=last, after=after, before=before, count=count
                )
                if limit:
                    if reverse:
                        args["pk__in"] = args["pk__in"][::-1][skip : skip + limit]
                    else:
                        args["pk__in"] = args["pk__in"][skip : skip + limit]
                elif skip:
                    args["pk__in"] = args["pk__in"][skip:]
                iterables = self.get_queryset(self.model, info, required_fields, **args)
                iterables = await sync_to_async(list)(iterables)
                list_length = len(iterables)
                if isinstance(info, GraphQLResolveInfo):
                    if not info.context:
                        info = info._replace(context=Context())
                    info.context.queryset = self.get_queryset(
                        self.model, info, required_fields, **args
                    )

        elif _root is not None:
            field_name = to_snake_case(info.field_name)
            items = getattr(_root, field_name, [])
            count = len(items)
            skip, limit, reverse = find_skip_and_limit(
                first=first, last=last, after=after, before=before, count=count
            )
            if limit:
                if reverse:
                    _base_query = items[::-1]
                    items = _base_query[skip : skip + limit]
                else:
                    _base_query = items
                    items = items[skip : skip + limit]
                has_next_page = (skip + limit) < len(_base_query) if requires_page_info else False
            elif skip:
                items = items[skip:]
            iterables = items
            iterables = await sync_to_async(list)(iterables)
            list_length = len(iterables)

        if requires_page_info and count:
            has_next_page = (
                True
                if (0 if limit is None else limit) + (0 if skip is None else skip) < count
                else False
            )
        has_previous_page = True if requires_page_info and skip else False

        if reverse:
            iterables = await sync_to_async(list)(iterables)
            iterables.reverse()
            skip = limit

        connection = connection_from_iterables(
            edges=iterables,
            start_offset=skip,
            has_previous_page=has_previous_page,
            has_next_page=has_next_page,
            connection_type=self.type,
            edge_type=self.type.Edge,
            pageinfo_type=graphene.PageInfo,
        )
        connection.iterable = iterables
        connection.list_length = list_length
        return connection

    async def chained_resolver(self, resolver, is_partial, root, info, **args):
        for key, value in dict(args).items():
            if value is None:
                del args[key]

        required_fields = list()

        for field in self.required_fields:
            if field in self.model._fields_ordered:
                required_fields.append(field)

        for field in get_query_fields(info):
            if to_snake_case(field) in self.model._fields_ordered:
                required_fields.append(to_snake_case(field))

        args_copy = args.copy()

        if not bool(args) or not is_partial:
            if isinstance(self.model, mongoengine.Document) or isinstance(
                self.model, mongoengine.base.metaclasses.TopLevelDocumentMetaclass
            ):
                connection_fields = [
                    field
                    for field in self.fields
                    if isinstance(self.fields[field], AsyncMongoengineConnectionField)
                ]

                def filter_connection(x):
                    return any(
                        [
                            connection_fields.__contains__(x),
                            self._type._meta.non_filter_fields.__contains__(x),
                        ]
                    )

                filterable_args = tuple(
                    filterfalse(filter_connection, list(self.model._fields_ordered))
                )
                for arg_name, arg in args.copy().items():
                    if arg_name not in filterable_args + tuple(self.filter_args.keys()):
                        args_copy.pop(arg_name)
                if isinstance(info, GraphQLResolveInfo):
                    if not info.context:
                        info = info._replace(context=Context())
                    info.context.queryset = self.get_queryset(
                        self.model, info, required_fields, **args_copy
                    )

            # XXX: Filter nested args
            resolved = resolver(root, info, **args)
            if isinstance(resolved, Coroutine):
                resolved = await resolved
            if resolved is not None:
                # if isinstance(resolved, Coroutine):
                #     resolved = await resolved
                if isinstance(resolved, list):
                    if resolved == list():
                        return resolved
                    elif not isinstance(resolved[0], DBRef):
                        return resolved
                    else:
                        return await self.default_resolver(root, info, required_fields, **args_copy)
                elif isinstance(resolved, QuerySet):
                    args.update(resolved._query)
                    args_copy = args.copy()
                    for arg_name, arg in args.copy().items():
                        if "." in arg_name or arg_name not in self.model._fields_ordered + (
                            "first",
                            "last",
                            "before",
                            "after",
                        ) + tuple(self.filter_args.keys()):
                            args_copy.pop(arg_name)
                            if arg_name == "_id" and isinstance(arg, dict):
                                operation = list(arg.keys())[0]
                                args_copy["pk" + operation.replace("$", "__")] = arg[operation]
                            if not isinstance(arg, ObjectId) and "." in arg_name:
                                if isinstance(arg, dict):
                                    operation = list(arg.keys())[0]
                                    args_copy[
                                        arg_name.replace(".", "__") + operation.replace("$", "__")
                                    ] = arg[operation]
                                else:
                                    args_copy[arg_name.replace(".", "__")] = arg
                            elif "." in arg_name and isinstance(arg, ObjectId):
                                args_copy[arg_name.replace(".", "__")] = arg
                        else:
                            operations = ["$lte", "$gte", "$ne", "$in"]
                            if isinstance(arg, dict) and any(op in arg for op in operations):
                                operation = list(arg.keys())[0]
                                args_copy[arg_name + operation.replace("$", "__")] = arg[operation]
                                del args_copy[arg_name]

                    return await self.default_resolver(
                        root, info, required_fields, resolved=resolved, **args_copy
                    )
                elif isinstance(resolved, Promise):
                    return resolved.value
                else:
                    return await resolved

        return await self.default_resolver(root, info, required_fields, **args)

    @classmethod
    async def connection_resolver(cls, resolver, connection_type, root, info, **args):
        if root:
            for key, value in root.__dict__.items():
                if value:
                    try:
                        setattr(root, key, from_global_id(value)[1])
                    except Exception:
                        pass

        iterable = await resolver(root=root, info=info, **args)

        if isinstance(connection_type, graphene.NonNull):
            connection_type = connection_type.of_type
        on_resolve = partial(cls.resolve_connection, connection_type, args)
        if Promise.is_thenable(iterable):
            iterable = Promise.resolve(iterable).then(on_resolve).value
        return on_resolve(iterable)



from __future__ import unicode_literals

import enum
import inspect
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Callable, Union

import mongoengine
from asgiref.sync import sync_to_async as asgiref_sync_to_async
from asgiref.sync import SyncToAsync
from graphene import Node
from graphene.utils.trim_docstring import trim_docstring
from graphql import FieldNode
from graphql_relay.connection.array_connection import offset_to_cursor


class ExecutorEnum(enum.Enum):
    ASYNC = enum.auto()
    SYNC = enum.auto()


def get_model_fields(model, excluding=None):
    excluding = excluding or []
    attributes = dict()
    for attr_name, attr in model._fields.items():
        if attr_name in excluding:
            continue
        attributes[attr_name] = attr
    return OrderedDict(sorted(attributes.items()))


def get_model_reference_fields(model, excluding=None):
    excluding = excluding or []
    attributes = dict()
    for attr_name, attr in model._fields.items():
        if attr_name in excluding or not isinstance(
            attr,
            (mongoengine.fields.ReferenceField, mongoengine.fields.LazyReferenceField),
        ):
            continue
        attributes[attr_name] = attr
    return attributes


def is_valid_mongoengine_model(model):
    return inspect.isclass(model) and (
        issubclass(model, mongoengine.Document) or issubclass(model, mongoengine.EmbeddedDocument)
    )


def import_single_dispatch():
    try:
        from functools import singledispatch
    except ImportError:
        singledispatch = None

    if not singledispatch:
        try:
            from singledispatch import singledispatch
        except ImportError:
            pass

    if not singledispatch:
        raise Exception(
            "It seems your python version does not include "
            "functools.singledispatch. Please install the 'singledispatch' "
            "package. More information here: "
            "https://pypi.python.org/pypi/singledispatch"
        )

    return singledispatch


# noqa
def get_type_for_document(schema, document):
    types = schema.types.values()
    for _type in types:
        type_document = hasattr(_type, "_meta") and getattr(_type._meta, "document", None)
        if document == type_document:
            return _type


def get_field_description(field, registry=None):
    """
    Common metadata includes verbose_name and help_text.

    http://docs.mongoengine.org/apireference.html#fields
    """
    parts = []
    if hasattr(field, "document_type"):
        doc = trim_docstring(field.document_type.__doc__)
        if doc:
            parts.append(doc)
    if hasattr(field, "verbose_name"):
        parts.append(field.verbose_name.title())
    if hasattr(field, "help_text"):
        parts.append(field.help_text)
    if field.db_field != field.name:
        name_format = "(%s)" if parts else "%s"
        parts.append(name_format % field.db_field)

    return "\n".join(parts)


def get_node_from_global_id(node, info, global_id):
    try:
        for interface in node._meta.interfaces:
            if issubclass(interface, Node):
                return interface.get_node_from_global_id(info, global_id)
    except AttributeError:
        return Node.get_node_from_global_id(info, global_id)


def collect_query_fields(node, fragments):
    """Recursively collects fields from the AST

    Args:
        node (dict): A node in the AST
        fragments (dict): Fragment definitions

    Returns:
        A dict mapping each field found, along with their sub fields.
        {
            'name': {},
            'image': {
                        'id': {},
                        'name': {},
                        'description': {}
                    },
            'slug': {}
        }
    """

    field = {}
    selection_set = None
    if isinstance(node, dict):
        selection_set = node.get("selection_set")
    else:
        selection_set = node.selection_set
    if selection_set:
        for leaf in selection_set.selections:
            if leaf.kind == "field":
                field.update({leaf.name.value: collect_query_fields(leaf, fragments)})
            elif leaf.kind == "fragment_spread":
                field.update(collect_query_fields(fragments[leaf.name.value], fragments))
            elif leaf.kind == "inline_fragment":
                field.update(
                    {leaf.type_condition.name.value: collect_query_fields(leaf, fragments)}
                )

    return field


def get_query_fields(info):
    """A convenience function to call collect_query_fields with info

    Args:
        info (ResolveInfo)

    Returns:
        dict: Returned from collect_query_fields
    """

    fragments = {}
    node = ast_to_dict(info.field_nodes[0])

    for name, value in info.fragments.items():
        fragments[name] = ast_to_dict(value)

    query = collect_query_fields(node, fragments)
    if "edges" in query:
        return query["edges"]["node"].keys()
    return query


def has_page_info(info):
    """A convenience function to call collect_query_fields with info
    for retrieving if page_info details are required

    Args:
        info (ResolveInfo)

    Returns:
        bool: True if it received pageinfo
    """

    fragments = {}
    if not info:
        return True  # Returning True if invalid info is provided
    node = ast_to_dict(info.field_nodes[0])
    for name, value in info.fragments.items():
        fragments[name] = ast_to_dict(value)

    query = collect_query_fields(node, fragments)
    return next((True for x in query.keys() if x.lower() == "pageinfo"), False)


def ast_to_dict(node, include_loc=False):
    if isinstance(node, FieldNode):
        d = {"kind": node.__class__.__name__}
        if hasattr(node, "keys"):
            for field in node.keys:
                d[field] = ast_to_dict(getattr(node, field), include_loc)

        if include_loc and hasattr(node, "loc") and node.loc:
            d["loc"] = {"start": node.loc.start, "end": node.loc.end}

        return d

    elif isinstance(node, list):
        return [ast_to_dict(item, include_loc) for item in node]

    return node


def find_skip_and_limit(first, last, after, before, count=None):
    reverse = False
    skip = 0
    limit = None
    if first is not None and after is not None:
        skip = after + 1
        limit = first
    elif first is not None and before is not None:
        if first >= before:
            limit = before - 1
        else:
            limit = first
    elif first is not None:
        skip = 0
        limit = first
    elif last is not None and before is not None:
        reverse = False
        if last >= before:
            limit = before
        else:
            limit = last
            skip = before - last
    elif last is not None and after is not None:
        if not count:
            raise ValueError("Count Missing")
        reverse = True
        if last + after < count:
            limit = last
        else:
            limit = count - after - 1
    elif last is not None:
        skip = 0
        limit = last
        reverse = True
    elif after is not None:
        skip = after + 1
    elif before is not None:
        limit = before
    return skip, limit, reverse


def connection_from_iterables(
    edges,
    start_offset,
    has_previous_page,
    has_next_page,
    connection_type,
    edge_type,
    pageinfo_type,
):
    edges_items = [
        edge_type(
            node=node,
            cursor=offset_to_cursor((0 if start_offset is None else start_offset) + i),
        )
        for i, node in enumerate(edges)
    ]

    first_edge_cursor = edges_items[0].cursor if edges_items else None
    last_edge_cursor = edges_items[-1].cursor if edges_items else None

    return connection_type(
        edges=edges_items,
        page_info=pageinfo_type(
            start_cursor=first_edge_cursor,
            end_cursor=last_edge_cursor,
            has_previous_page=has_previous_page,
            has_next_page=has_next_page,
        ),
    )


def sync_to_async(
    func: Callable = None,
    thread_sensitive: bool = False,
    executor: Any = None,  # noqa
) -> Union[SyncToAsync, Callable[[Callable[..., Any]], SyncToAsync]]:
    """
    Wrapper over sync_to_async from asgiref.sync
    Defaults to thread insensitive with ThreadPoolExecutor of n workers
    Args:
        func:
            Function to be converted to coroutine
        thread_sensitive:
            If the operation is thread sensitive and should run in synchronous thread
        executor:
            Threadpool executor, if thread_sensitive=False

    Returns:
        coroutine version of func
    """
    if executor is None:
        executor = ThreadPoolExecutor()
    return asgiref_sync_to_async(func=func, thread_sensitive=thread_sensitive, executor=executor)



import graphene
from graphene_mongo.fields import MongoengineConnectionField
from .types import CategoryType, BookmarkType


class Query(graphene.ObjectType):
    categories = MongoengineConnectionField(CategoryType)
    bookmarks = MongoengineConnectionField(BookmarkType)


schema = graphene.Schema(query=Query, types=[CategoryType, BookmarkType])



import json
import falcon
from .schema import schema


def set_graphql_allow_header(req: falcon.Request, resp: falcon.Response, resource: object):
    resp.set_header("Allow", "GET, POST, OPTIONS")


class HelloWorldResource:
    def on_get(self, req, resp):
        name = "Hello World!"
        resp.status = falcon.HTTP_200
        resp.body = json.dumps({"respone": name, "status": resp.status})

    def on_post(self, req, resp):
        pass


@falcon.after(set_graphql_allow_header)
class GraphQLResource:
    def on_get(self, req, resp):
        query = req.params["query"]
        result = await schema.execute_async(query)

        if result.data:
            data_ret = {"data": result.data}
            resp.status = falcon.HTTP_200
            resp.body = json.dumps(data_ret, separators=(",", ":"))

    def on_post(self, req, resp):
        query = req.params["query"]
        result = await schema.execute_async(query)
        if result.data:
            data_ret = {"data": result.data}
            resp.status = falcon.HTTP_200
            resp.body = json.dumps(data_ret, separators=(",", ":"))



from mongoengine import Document
from mongoengine.fields import (
    FloatField,
    StringField,
    ListField,
    URLField,
    ObjectIdField,
)


class Shop(Document):
    meta = {"collection": "shop"}
    ID = ObjectIdField()
    name = StringField()
    address = StringField()
    website = URLField()


class Bike(Document):
    meta = {"collection": "bike"}
    ID = ObjectIdField()
    name = StringField()
    brand = StringField()
    year = StringField()
    size = ListField(StringField())
    wheel_size = FloatField()
    type = StringField()



import asyncio
import sys

import graphene
import mongoengine

from graphene.types.json import JSONString
from graphene.utils.str_converters import to_snake_case, to_camel_case
from mongoengine.base import get_document, LazyReference
from . import advanced_types
from .utils import (
    import_single_dispatch,
    get_field_description,
    get_query_fields,
    ExecutorEnum,
    sync_to_async,
)
from concurrent.futures import ThreadPoolExecutor, as_completed

singledispatch = import_single_dispatch()


class MongoEngineConversionError(Exception):
    pass


@singledispatch
def convert_mongoengine_field(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    raise MongoEngineConversionError(
        "Don't know how to convert the MongoEngine field %s (%s)" % (field, field.__class__)
    )


@convert_mongoengine_field.register(mongoengine.EmailField)
@convert_mongoengine_field.register(mongoengine.StringField)
@convert_mongoengine_field.register(mongoengine.URLField)
def convert_field_to_string(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    return graphene.String(
        description=get_field_description(field, registry), required=field.required
    )


@convert_mongoengine_field.register(mongoengine.UUIDField)
@convert_mongoengine_field.register(mongoengine.ObjectIdField)
def convert_field_to_id(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    return graphene.ID(description=get_field_description(field, registry), required=field.required)


@convert_mongoengine_field.register(mongoengine.IntField)
@convert_mongoengine_field.register(mongoengine.LongField)
@convert_mongoengine_field.register(mongoengine.SequenceField)
def convert_field_to_int(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    return graphene.Int(description=get_field_description(field, registry), required=field.required)


@convert_mongoengine_field.register(mongoengine.BooleanField)
def convert_field_to_boolean(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    return graphene.Boolean(
        description=get_field_description(field, registry), required=field.required
    )


@convert_mongoengine_field.register(mongoengine.FloatField)
def convert_field_to_float(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    return graphene.Float(
        description=get_field_description(field, registry), required=field.required
    )


@convert_mongoengine_field.register(mongoengine.Decimal128Field)
@convert_mongoengine_field.register(mongoengine.DecimalField)
def convert_field_to_decimal(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    return graphene.Decimal(
        description=get_field_description(field, registry), required=field.required
    )


@convert_mongoengine_field.register(mongoengine.DateTimeField)
def convert_field_to_datetime(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    return graphene.DateTime(
        description=get_field_description(field, registry), required=field.required
    )


@convert_mongoengine_field.register(mongoengine.DateField)
def convert_field_to_date(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    return graphene.Date(
        description=get_field_description(field, registry), required=field.required
    )


@convert_mongoengine_field.register(mongoengine.DictField)
@convert_mongoengine_field.register(mongoengine.MapField)
def convert_field_to_jsonstring(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    return JSONString(description=get_field_description(field, registry), required=field.required)


@convert_mongoengine_field.register(mongoengine.PointField)
def convert_point_to_field(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    return graphene.Field(
        advanced_types.PointFieldType,
        description=get_field_description(field, registry),
        required=field.required,
    )


@convert_mongoengine_field.register(mongoengine.PolygonField)
def convert_polygon_to_field(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    return graphene.Field(
        advanced_types.PolygonFieldType,
        description=get_field_description(field, registry),
        required=field.required,
    )


@convert_mongoengine_field.register(mongoengine.MultiPolygonField)
def convert_multipolygon_to_field(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    return graphene.Field(
        advanced_types.MultiPolygonFieldType,
        description=get_field_description(field, registry),
        required=field.required,
    )


@convert_mongoengine_field.register(mongoengine.FileField)
def convert_file_to_field(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    return graphene.Field(
        advanced_types.FileFieldType,
        description=get_field_description(field, registry),
        required=field.required,
    )


@convert_mongoengine_field.register(mongoengine.ListField)
@convert_mongoengine_field.register(mongoengine.EmbeddedDocumentListField)
@convert_mongoengine_field.register(mongoengine.GeoPointField)
def convert_field_to_list(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    base_type = convert_mongoengine_field(field.field, registry=registry, executor=executor)
    if isinstance(base_type, graphene.Field):
        if isinstance(field.field, mongoengine.GenericReferenceField):

            def get_reference_objects(*args, **kwargs):
                document = get_document(args[0][0])
                document_field = mongoengine.ReferenceField(document)
                document_field = convert_mongoengine_field(document_field, registry)
                document_field_type = document_field.get_type().type
                queried_fields = list()
                filter_args = list()
                if document_field_type._meta.filter_fields:
                    for key, values in document_field_type._meta.filter_fields.items():
                        for each in values:
                            filter_args.append(key + "__" + each)
                for each in get_query_fields(args[0][3][0])[document_field_type._meta.name].keys():
                    item = to_snake_case(each)
                    if item in document._fields_ordered + tuple(filter_args):
                        queried_fields.append(item)
                return (
                    document.objects()
                    .no_dereference()
                    .only(*set(list(document_field_type._meta.required_fields) + queried_fields))
                    .filter(pk__in=args[0][1])
                )

            def get_non_querying_object(*args, **kwargs):
                model = get_document(args[0][0])
                return [model(pk=each) for each in args[0][1]]

            def reference_resolver(root, *args, **kwargs):
                to_resolve = getattr(root, field.name or field.db_name)
                if to_resolve:
                    choice_to_resolve = dict()
                    querying_union_types = list(get_query_fields(args[0]).keys())
                    if "__typename" in querying_union_types:
                        querying_union_types.remove("__typename")
                    to_resolve_models = list()
                    for each in querying_union_types:
                        if executor == ExecutorEnum.SYNC:
                            to_resolve_models.append(registry._registry_string_map[each])
                        else:
                            to_resolve_models.append(registry._registry_async_string_map[each])
                    to_resolve_object_ids = list()
                    for each in to_resolve:
                        if isinstance(each, LazyReference):
                            to_resolve_object_ids.append(each.pk)
                            model = each.document_type._class_name
                            if model not in choice_to_resolve:
                                choice_to_resolve[model] = list()
                            choice_to_resolve[model].append(each.pk)
                        else:
                            to_resolve_object_ids.append(each["_ref"].id)
                            if each["_cls"] not in choice_to_resolve:
                                choice_to_resolve[each["_cls"]] = list()
                            choice_to_resolve[each["_cls"]].append(each["_ref"].id)
                    pool = ThreadPoolExecutor(5)
                    futures = list()
                    for model, object_id_list in choice_to_resolve.items():
                        if model in to_resolve_models:
                            futures.append(
                                pool.submit(
                                    get_reference_objects,
                                    (model, object_id_list, registry, args),
                                )
                            )
                        else:
                            futures.append(
                                pool.submit(
                                    get_non_querying_object,
                                    (model, object_id_list, registry, args),
                                )
                            )
                    result = list()
                    for x in as_completed(futures):
                        result += x.result()
                    result_object_ids = list()
                    for each in result:
                        result_object_ids.append(each.id)
                    ordered_result = list()
                    for each in to_resolve_object_ids:
                        ordered_result.append(result[result_object_ids.index(each)])
                    return ordered_result
                return None

            async def get_reference_objects_async(*args, **kwargs):
                document = get_document(args[0])
                document_field = mongoengine.ReferenceField(document)
                document_field = convert_mongoengine_field(
                    document_field, registry, executor=ExecutorEnum.ASYNC
                )
                document_field_type = document_field.get_type().type
                queried_fields = list()
                filter_args = list()
                if document_field_type._meta.filter_fields:
                    for key, values in document_field_type._meta.filter_fields.items():
                        for each in values:
                            filter_args.append(key + "__" + each)
                for each in get_query_fields(args[3][0])[document_field_type._meta.name].keys():
                    item = to_snake_case(each)
                    if item in document._fields_ordered + tuple(filter_args):
                        queried_fields.append(item)
                return await sync_to_async(list)(
                    document.objects()
                    .no_dereference()
                    .only(*set(list(document_field_type._meta.required_fields) + queried_fields))
                    .filter(pk__in=args[1])
                )

            async def get_non_querying_object_async(*args, **kwargs):
                model = get_document(args[0])
                return [model(pk=each) for each in args[1]]

            async def reference_resolver_async(root, *args, **kwargs):
                to_resolve = getattr(root, field.name or field.db_name)
                if to_resolve:
                    choice_to_resolve = dict()
                    querying_union_types = list(get_query_fields(args[0]).keys())
                    if "__typename" in querying_union_types:
                        querying_union_types.remove("__typename")
                    to_resolve_models = list()
                    for each in querying_union_types:
                        if executor == ExecutorEnum.SYNC:
                            to_resolve_models.append(registry._registry_string_map[each])
                        else:
                            to_resolve_models.append(registry._registry_async_string_map[each])
                    to_resolve_object_ids = list()
                    for each in to_resolve:
                        if isinstance(each, LazyReference):
                            to_resolve_object_ids.append(each.pk)
                            model = each.document_type._class_name
                            if model not in choice_to_resolve:
                                choice_to_resolve[model] = list()
                            choice_to_resolve[model].append(each.pk)
                        else:
                            to_resolve_object_ids.append(each["_ref"].id)
                            if each["_cls"] not in choice_to_resolve:
                                choice_to_resolve[each["_cls"]] = list()
                            choice_to_resolve[each["_cls"]].append(each["_ref"].id)
                    loop = asyncio.get_event_loop()
                    tasks = []
                    for model, object_id_list in choice_to_resolve.items():
                        if model in to_resolve_models:
                            task = loop.create_task(
                                get_reference_objects_async(model, object_id_list, registry, args)
                            )
                        else:
                            task = loop.create_task(
                                get_non_querying_object_async(model, object_id_list, registry, args)
                            )
                        tasks.append(task)
                    result = await asyncio.gather(*tasks)
                    result_object = {}
                    for items in result:
                        for item in items:
                            result_object[item.id] = item
                    ordered_result = list()
                    for each in to_resolve_object_ids:
                        ordered_result.append(result_object[each])
                    return ordered_result
                return None

            return graphene.List(
                base_type._type,
                description=get_field_description(field, registry),
                required=field.required,
                resolver=reference_resolver
                if executor == ExecutorEnum.SYNC
                else reference_resolver_async,
            )
        return graphene.List(
            base_type._type,
            description=get_field_description(field, registry),
            required=field.required,
        )
    if isinstance(base_type, (graphene.Dynamic)):
        base_type = base_type.get_type()
        if base_type is None:
            return
        base_type = base_type._type

    if graphene.is_node(base_type):
        return base_type._meta.connection_field_class(base_type)

    # Non-relationship field
    relations = (mongoengine.ReferenceField, mongoengine.EmbeddedDocumentField)
    if not isinstance(base_type, (graphene.List, graphene.NonNull)) and not isinstance(
        field.field, relations
    ):
        base_type = type(base_type)

    return graphene.List(
        base_type,
        description=get_field_description(field, registry),
        required=field.required,
    )


@convert_mongoengine_field.register(mongoengine.GenericEmbeddedDocumentField)
@convert_mongoengine_field.register(mongoengine.GenericReferenceField)
def convert_field_to_union(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    _types = []
    for choice in field.choices:
        if isinstance(field, mongoengine.GenericReferenceField):
            _field = mongoengine.ReferenceField(get_document(choice))
        elif isinstance(field, mongoengine.GenericEmbeddedDocumentField):
            _field = mongoengine.EmbeddedDocumentField(choice)

        _field = convert_mongoengine_field(_field, registry, executor=executor)
        _type = _field.get_type()
        if _type:
            _types.append(_type.type)
        else:
            # TODO: Register type auto-matically here.
            pass

    if len(_types) == 0:
        return None

    field_name = field.db_field
    if field_name is None:
        # Get db_field name from parent mongo_field
        for db_field_name, _mongo_parent_field in field.owner_document._fields.items():
            if hasattr(_mongo_parent_field, "field") and _mongo_parent_field.field == field:
                field_name = db_field_name
                break

    name = to_camel_case(
        "{}_{}_union_type".format(
            field._owner_document.__name__,
            field_name,
        )
    )
    Meta = type("Meta", (object,), {"types": tuple(_types)})
    _union = type(name, (graphene.Union,), {"Meta": Meta})

    def reference_resolver(root, *args, **kwargs):
        de_referenced = getattr(root, field.name or field.db_name)
        if de_referenced:
            document = get_document(de_referenced["_cls"])
            document_field = mongoengine.ReferenceField(document)
            document_field = convert_mongoengine_field(document_field, registry, executor=executor)
            _type = document_field.get_type().type
            filter_args = list()
            if _type._meta.filter_fields:
                for key, values in _type._meta.filter_fields.items():
                    for each in values:
                        filter_args.append(key + "__" + each)
            querying_types = list(get_query_fields(args[0]).keys())
            if _type.__name__ in querying_types:
                queried_fields = list()
                for each in get_query_fields(args[0])[_type._meta.name].keys():
                    item = to_snake_case(each)
                    if item in document._fields_ordered + tuple(filter_args):
                        queried_fields.append(item)
                return (
                    document.objects()
                    .no_dereference()
                    .only(*list(set(list(_type._meta.required_fields) + queried_fields)))
                    .get(pk=de_referenced["_ref"].id)
                )
            return document()
        return None

    def lazy_reference_resolver(root, *args, **kwargs):
        document = getattr(root, field.name or field.db_name)
        if document:
            if document._cached_doc:
                return document._cached_doc
            queried_fields = list()
            document_field_type = registry.get_type_for_model(
                document.document_type, executor=executor
            )
            querying_types = list(get_query_fields(args[0]).keys())
            filter_args = list()
            if document_field_type._meta.filter_fields:
                for key, values in document_field_type._meta.filter_fields.items():
                    for each in values:
                        filter_args.append(key + "__" + each)
            if document_field_type._meta.name in querying_types:
                for each in get_query_fields(args[0])[document_field_type._meta.name].keys():
                    item = to_snake_case(each)
                    if item in document.document_type._fields_ordered + tuple(filter_args):
                        queried_fields.append(item)
                _type = registry.get_type_for_model(document.document_type, executor=executor)
                return (
                    document.document_type.objects()
                    .no_dereference()
                    .only(*(set((list(_type._meta.required_fields) + queried_fields))))
                    .get(pk=document.pk)
                )
            return document.document_type()
        return None

    async def reference_resolver_async(root, *args, **kwargs):
        de_referenced = getattr(root, field.name or field.db_name)
        if de_referenced:
            document = get_document(de_referenced["_cls"])
            document_field = mongoengine.ReferenceField(document)
            document_field = convert_mongoengine_field(
                document_field, registry, executor=ExecutorEnum.ASYNC
            )
            _type = document_field.get_type().type
            filter_args = list()
            if _type._meta.filter_fields:
                for key, values in _type._meta.filter_fields.items():
                    for each in values:
                        filter_args.append(key + "__" + each)
            querying_types = list(get_query_fields(args[0]).keys())
            if _type.__name__ in querying_types:
                queried_fields = list()
                for each in get_query_fields(args[0])[_type._meta.name].keys():
                    item = to_snake_case(each)
                    if item in document._fields_ordered + tuple(filter_args):
                        queried_fields.append(item)
                return await sync_to_async(
                    document.objects()
                    .no_dereference()
                    .only(*list(set(list(_type._meta.required_fields) + queried_fields)))
                    .get
                )(pk=de_referenced["_ref"].id)
            return await sync_to_async(document)()
        return None

    async def lazy_reference_resolver_async(root, *args, **kwargs):
        document = getattr(root, field.name or field.db_name)
        if document:
            if document._cached_doc:
                return document._cached_doc
            queried_fields = list()
            document_field_type = registry.get_type_for_model(
                document.document_type, executor=executor
            )
            querying_types = list(get_query_fields(args[0]).keys())
            filter_args = list()
            if document_field_type._meta.filter_fields:
                for key, values in document_field_type._meta.filter_fields.items():
                    for each in values:
                        filter_args.append(key + "__" + each)
            if document_field_type._meta.name in querying_types:
                for each in get_query_fields(args[0])[document_field_type._meta.name].keys():
                    item = to_snake_case(each)
                    if item in document.document_type._fields_ordered + tuple(filter_args):
                        queried_fields.append(item)
                _type = registry.get_type_for_model(document.document_type, executor=executor)
                return await sync_to_async(
                    document.document_type.objects()
                    .no_dereference()
                    .only(*(set((list(_type._meta.required_fields) + queried_fields))))
                    .get
                )(pk=document.pk)
            return await sync_to_async(document.document_type)()
        return None

    if isinstance(field, mongoengine.GenericLazyReferenceField):
        field_resolver = None
        required = False
        if field.db_field is not None:
            required = field.required
            resolver_function = getattr(
                registry.get_type_for_model(field.owner_document, executor=executor),
                "resolve_" + field.db_field,
                None,
            )
            if resolver_function and callable(resolver_function):
                field_resolver = resolver_function
        return graphene.Field(
            _union,
            resolver=field_resolver
            if field_resolver
            else (
                lazy_reference_resolver
                if executor == ExecutorEnum.SYNC
                else lazy_reference_resolver_async
            ),
            description=get_field_description(field, registry),
            required=required,
        )

    elif isinstance(field, mongoengine.GenericReferenceField):
        field_resolver = None
        required = False
        if field.db_field is not None:
            required = field.required
            resolver_function = getattr(
                registry.get_type_for_model(field.owner_document, executor=executor),
                "resolve_" + field.db_field,
                None,
            )
            if resolver_function and callable(resolver_function):
                field_resolver = resolver_function
        return graphene.Field(
            _union,
            resolver=field_resolver
            if field_resolver
            else (
                reference_resolver if executor == ExecutorEnum.SYNC else reference_resolver_async
            ),
            description=get_field_description(field, registry),
            required=required,
        )

    return graphene.Field(_union)


@convert_mongoengine_field.register(mongoengine.EmbeddedDocumentField)
@convert_mongoengine_field.register(mongoengine.ReferenceField)
@convert_mongoengine_field.register(mongoengine.CachedReferenceField)
def convert_field_to_dynamic(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    model = field.document_type

    def reference_resolver(root, *args, **kwargs):
        document = getattr(root, field.name or field.db_name)
        if document:
            queried_fields = list()
            _type = registry.get_type_for_model(field.document_type, executor=executor)
            filter_args = list()
            if _type._meta.filter_fields:
                for key, values in _type._meta.filter_fields.items():
                    for each in values:
                        filter_args.append(key + "__" + each)
            for each in get_query_fields(args[0]).keys():
                item = to_snake_case(each)
                if item in field.document_type._fields_ordered + tuple(filter_args):
                    queried_fields.append(item)
            return (
                field.document_type.objects()
                .no_dereference()
                .only(*(set(list(_type._meta.required_fields) + queried_fields)))
                .get(pk=document.id)
            )
        return None

    def cached_reference_resolver(root, *args, **kwargs):
        if field:
            queried_fields = list()
            _type = registry.get_type_for_model(field.document_type, executor=executor)
            filter_args = list()
            if _type._meta.filter_fields:
                for key, values in _type._meta.filter_fields.items():
                    for each in values:
                        filter_args.append(key + "__" + each)
            for each in get_query_fields(args[0]).keys():
                item = to_snake_case(each)
                if item in field.document_type._fields_ordered + tuple(filter_args):
                    queried_fields.append(item)
            return (
                field.document_type.objects()
                .no_dereference()
                .only(*(set(list(_type._meta.required_fields) + queried_fields)))
                .get(pk=getattr(root, field.name or field.db_name))
            )
        return None

    async def reference_resolver_async(root, *args, **kwargs):
        document = getattr(root, field.name or field.db_name)
        if document:
            queried_fields = list()
            _type = registry.get_type_for_model(field.document_type, executor=executor)
            filter_args = list()
            if _type._meta.filter_fields:
                for key, values in _type._meta.filter_fields.items():
                    for each in values:
                        filter_args.append(key + "__" + each)
            for each in get_query_fields(args[0]).keys():
                item = to_snake_case(each)
                if item in field.document_type._fields_ordered + tuple(filter_args):
                    queried_fields.append(item)
            return await sync_to_async(
                field.document_type.objects()
                .no_dereference()
                .only(*(set(list(_type._meta.required_fields) + queried_fields)))
                .get
            )(pk=document.id)
        return None

    async def cached_reference_resolver_async(root, *args, **kwargs):
        if field:
            queried_fields = list()
            _type = registry.get_type_for_model(field.document_type, executor=executor)
            filter_args = list()
            if _type._meta.filter_fields:
                for key, values in _type._meta.filter_fields.items():
                    for each in values:
                        filter_args.append(key + "__" + each)
            for each in get_query_fields(args[0]).keys():
                item = to_snake_case(each)
                if item in field.document_type._fields_ordered + tuple(filter_args):
                    queried_fields.append(item)
            return await sync_to_async(
                field.document_type.objects()
                .no_dereference()
                .only(*(set(list(_type._meta.required_fields) + queried_fields)))
                .get
            )(pk=getattr(root, field.name or field.db_name))
        return None

    def dynamic_type():
        _type = registry.get_type_for_model(model, executor=executor)
        if not _type:
            return None
        if isinstance(field, mongoengine.EmbeddedDocumentField):
            return graphene.Field(
                _type,
                description=get_field_description(field, registry),
                required=field.required,
            )
        field_resolver = None
        required = False
        if field.db_field is not None:
            required = field.required
            resolver_function = getattr(
                registry.get_type_for_model(field.owner_document, executor=executor),
                "resolve_" + field.db_field,
                None,
            )
            if resolver_function and callable(resolver_function):
                field_resolver = resolver_function
        if isinstance(field, mongoengine.ReferenceField):
            return graphene.Field(
                _type,
                resolver=field_resolver
                if field_resolver
                else (
                    reference_resolver
                    if executor == ExecutorEnum.SYNC
                    else reference_resolver_async
                ),
                description=get_field_description(field, registry),
                required=required,
            )
        else:
            return graphene.Field(
                _type,
                resolver=field_resolver
                if field_resolver
                else (
                    cached_reference_resolver
                    if executor == ExecutorEnum.SYNC
                    else cached_reference_resolver_async
                ),
                description=get_field_description(field, registry),
                required=required,
            )

    return graphene.Dynamic(dynamic_type)


@convert_mongoengine_field.register(mongoengine.LazyReferenceField)
def convert_lazy_field_to_dynamic(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
    model = field.document_type

    def lazy_resolver(root, *args, **kwargs):
        document = getattr(root, field.name or field.db_name)
        if document:
            if document._cached_doc:
                return document._cached_doc
            queried_fields = list()
            _type = registry.get_type_for_model(document.document_type, executor=executor)
            filter_args = list()
            if _type._meta.filter_fields:
                for key, values in _type._meta.filter_fields.items():
                    for each in values:
                        filter_args.append(key + "__" + each)
            for each in get_query_fields(args[0]).keys():
                item = to_snake_case(each)
                if item in document.document_type._fields_ordered + tuple(filter_args):
                    queried_fields.append(item)
            return (
                document.document_type.objects()
                .no_dereference()
                .only(*(set((list(_type._meta.required_fields) + queried_fields))))
                .get(pk=document.pk)
            )
        return None

    async def lazy_resolver_async(root, *args, **kwargs):
        document = getattr(root, field.name or field.db_name)
        if document:
            if document._cached_doc:
                return document._cached_doc
            queried_fields = list()
            _type = registry.get_type_for_model(document.document_type, executor=executor)
            filter_args = list()
            if _type._meta.filter_fields:
                for key, values in _type._meta.filter_fields.items():
                    for each in values:
                        filter_args.append(key + "__" + each)
            for each in get_query_fields(args[0]).keys():
                item = to_snake_case(each)
                if item in document.document_type._fields_ordered + tuple(filter_args):
                    queried_fields.append(item)
            return await sync_to_async(
                document.document_type.objects()
                .no_dereference()
                .only(*(set((list(_type._meta.required_fields) + queried_fields))))
                .get
            )(pk=document.pk)
        return None

    def dynamic_type():
        _type = registry.get_type_for_model(model, executor=executor)
        if not _type:
            return None
        field_resolver = None
        required = False
        if field.db_field is not None:
            required = field.required
            resolver_function = getattr(
                registry.get_type_for_model(field.owner_document, executor=executor),
                "resolve_" + field.db_field,
                None,
            )
            if resolver_function and callable(resolver_function):
                field_resolver = resolver_function
        return graphene.Field(
            _type,
            resolver=field_resolver
            if field_resolver
            else (lazy_resolver if executor == ExecutorEnum.SYNC else lazy_resolver_async),
            description=get_field_description(field, registry),
            required=required,
        )

    return graphene.Dynamic(dynamic_type)


if sys.version_info >= (3, 6):

    @convert_mongoengine_field.register(mongoengine.EnumField)
    def convert_field_to_enum(field, registry=None, executor: ExecutorEnum = ExecutorEnum.SYNC):
        if not registry.check_enum_already_exist(field._enum_cls):
            registry.register_enum(field._enum_cls)
        _type = registry.get_type_for_enum(field._enum_cls)
        return graphene.Field(
            _type,
            description=get_field_description(field, registry),
            required=field.required,
        )



from . import models
from ..types import (
    MongoengineObjectType,
    MongoengineInterfaceType,
    MongoengineInputType,
)
from graphene.types.union import Union


class PublisherType(MongoengineObjectType):
    class Meta:
        model = models.Publisher


class EditorType(MongoengineObjectType):
    class Meta:
        model = models.Editor


class ArticleType(MongoengineObjectType):
    class Meta:
        model = models.Article


class EmbeddedArticleType(MongoengineObjectType):
    class Meta:
        model = models.EmbeddedArticle


class PlayerType(MongoengineObjectType):
    class Meta:
        model = models.Player


class ReporterType(MongoengineObjectType):
    class Meta:
        model = models.Reporter


class ParentType(MongoengineObjectType):
    class Meta:
        model = models.Parent


class ParentInterface(MongoengineInterfaceType):
    class Meta:
        model = models.Parent
        exclude_fields = ["loc"]


class ChildType(MongoengineObjectType):
    class Meta:
        model = models.Child
        interfaces = (ParentInterface,)


class AnotherChildType(MongoengineObjectType):
    class Meta:
        model = models.AnotherChild
        interfaces = (ParentInterface,)


class ChildUnionType(Union):
    class Meta:
        types = (ChildType, AnotherChildType)
        interfaces = (ParentInterface,)


class CellTowerType(MongoengineObjectType):
    class Meta:
        model = models.CellTower


class ProfessorMetadataType(MongoengineObjectType):
    class Meta:
        model = models.ProfessorMetadata


class ProfessorVectorType(MongoengineObjectType):
    class Meta:
        model = models.ProfessorVector


class ArticleInput(MongoengineInputType):
    class Meta:
        model = models.Article
        only_fields = ["headline"]


class EditorInput(MongoengineInputType):
    class Meta:
        model = models.Editor
        only_fields = ["first_name", "last_name"]
        # allow providing only one of those ! Even None...
        non_required_fields = ["first_name", "last_name"]



import graphene
from graphene.relay import Node
from graphene_mongo.fields import MongoengineConnectionField
from .models import Shop
from .types import BikeType, ShopType
from .mutations import CreateBikeMutation, UpdateBikeMutation, DeleteBikeMutation


class Mutations(graphene.ObjectType):
    create_bike = CreateBikeMutation.Field()
    update_bike = UpdateBikeMutation.Field()
    delete_bike = DeleteBikeMutation.Field()


class Query(graphene.ObjectType):
    node = Node.Field()
    bikes = MongoengineConnectionField(BikeType)
    shop_list = graphene.List(ShopType)

    def resolve_shop_list(self, info):
        return Shop.objects.all()


schema = graphene.Schema(query=Query, mutation=Mutations, types=[BikeType, ShopType])



import graphene
import mongoengine
from graphene import InputObjectType
from graphene.relay import Connection, Node
from graphene.types.interface import Interface, InterfaceOptions
from graphene.types.objecttype import ObjectType, ObjectTypeOptions
from graphene.types.utils import yank_fields_from_attrs
from graphene.utils.str_converters import to_snake_case

from graphene_mongo import AsyncMongoengineConnectionField
from .registry import Registry, get_global_async_registry, get_inputs_async_registry
from .types import construct_fields, construct_self_referenced_fields
from .utils import ExecutorEnum, get_query_fields, is_valid_mongoengine_model, sync_to_async


def create_graphene_generic_class_async(object_type, option_type):
    class AsyncMongoengineGenericObjectTypeOptions(option_type):
        model = None
        registry = None  # type: Registry
        connection = None
        filter_fields = ()
        non_required_fields = ()
        order_by = None

    class AsyncGrapheneMongoengineGenericType(object_type):
        @classmethod
        def __init_subclass_with_meta__(
            cls,
            model=None,
            registry=None,
            skip_registry=False,
            only_fields=(),
            required_fields=(),
            exclude_fields=(),
            non_required_fields=(),
            filter_fields=None,
            non_filter_fields=(),
            connection=None,
            connection_class=None,
            use_connection=None,
            connection_field_class=None,
            interfaces=(),
            _meta=None,
            order_by=None,
            **options,
        ):
            assert is_valid_mongoengine_model(model), (
                "The attribute model in {}.Meta must be a valid Mongoengine Model. "
                'Received "{}" instead.'
            ).format(cls.__name__, type(model))

            if not registry:
                # input objects shall be registred in a separated registry
                if issubclass(cls, InputObjectType):
                    registry = get_inputs_async_registry()
                else:
                    registry = get_global_async_registry()

            assert isinstance(registry, Registry), (
                "The attribute registry in {}.Meta needs to be an instance of "
                'Registry({}), received "{}".'
            ).format(object_type, cls.__name__, registry)
            converted_fields, self_referenced = construct_fields(
                model,
                registry,
                only_fields,
                exclude_fields,
                non_required_fields,
                ExecutorEnum.ASYNC,
            )
            mongoengine_fields = yank_fields_from_attrs(converted_fields, _as=graphene.Field)
            if use_connection is None and interfaces:
                use_connection = any((issubclass(interface, Node) for interface in interfaces))

            if use_connection and not connection:
                # We create the connection automatically
                if not connection_class:
                    connection_class = Connection

                connection = connection_class.create_type(
                    "{}Connection".format(options.get("name") or cls.__name__), node=cls
                )

            if connection is not None:
                assert issubclass(connection, Connection), (
                    "The attribute connection in {}.Meta must be of type Connection. "
                    'Received "{}" instead.'
                ).format(cls.__name__, type(connection))

            if connection_field_class is not None:
                assert issubclass(connection_field_class, graphene.ConnectionField), (
                    "The attribute connection_field_class in {}.Meta must be of type graphene.ConnectionField. "
                    'Received "{}" instead.'
                ).format(cls.__name__, type(connection_field_class))
            else:
                connection_field_class = AsyncMongoengineConnectionField

            if _meta:
                assert isinstance(_meta, AsyncMongoengineGenericObjectTypeOptions), (
                    "_meta must be an instance of AsyncMongoengineGenericObjectTypeOptions, "
                    "received {}"
                ).format(_meta.__class__)
            else:
                _meta = AsyncMongoengineGenericObjectTypeOptions(option_type)

            _meta.model = model
            _meta.registry = registry
            _meta.fields = mongoengine_fields
            _meta.filter_fields = filter_fields
            _meta.non_filter_fields = non_filter_fields
            _meta.connection = connection
            _meta.connection_field_class = connection_field_class
            # Save them for later
            _meta.only_fields = only_fields
            _meta.required_fields = required_fields
            _meta.exclude_fields = exclude_fields
            _meta.non_required_fields = non_required_fields
            _meta.order_by = order_by

            super(AsyncGrapheneMongoengineGenericType, cls).__init_subclass_with_meta__(
                _meta=_meta, interfaces=interfaces, **options
            )

            if not skip_registry:
                registry.register(cls)
                # Notes: Take care list of self-reference fields.
                converted_fields = construct_self_referenced_fields(
                    self_referenced, registry, ExecutorEnum.ASYNC
                )
                if converted_fields:
                    mongoengine_fields = yank_fields_from_attrs(
                        converted_fields, _as=graphene.Field
                    )
                    cls._meta.fields.update(mongoengine_fields)
                    registry.register(cls)

        @classmethod
        def rescan_fields(cls):
            """Attempts to rescan fields and will insert any not converted initially"""

            converted_fields, self_referenced = construct_fields(
                cls._meta.model,
                cls._meta.registry,
                cls._meta.only_fields,
                cls._meta.exclude_fields,
                cls._meta.non_required_fields,
                ExecutorEnum.ASYNC,
            )

            mongoengine_fields = yank_fields_from_attrs(converted_fields, _as=graphene.Field)

            # The initial scan should take precedence
            for field in mongoengine_fields:
                if field not in cls._meta.fields:
                    cls._meta.fields.update({field: mongoengine_fields[field]})
            # Self-referenced fields can't change between scans!

        @classmethod
        def is_type_of(cls, root, info):
            if isinstance(root, cls):
                return True
            # XXX: Take care FileField
            if isinstance(root, mongoengine.GridFSProxy):
                return True
            if not is_valid_mongoengine_model(type(root)):
                raise Exception(('Received incompatible instance "{}".').format(root))
            return isinstance(root, cls._meta.model)

        @classmethod
        async def get_node(cls, info, id):
            required_fields = list()
            for field in cls._meta.required_fields:
                if field in cls._meta.model._fields_ordered:
                    required_fields.append(field)
            queried_fields = get_query_fields(info)
            if cls._meta.name in queried_fields:
                queried_fields = queried_fields[cls._meta.name]
            for field in queried_fields:
                if to_snake_case(field) in cls._meta.model._fields_ordered:
                    required_fields.append(to_snake_case(field))
            required_fields = list(set(required_fields))
            return await sync_to_async(
                cls._meta.model.objects.no_dereference().only(*required_fields).get
            )(pk=id)

        def resolve_id(self, info):
            return str(self.id)

    return AsyncGrapheneMongoengineGenericType, AsyncMongoengineGenericObjectTypeOptions


(
    AsyncMongoengineObjectType,
    AsyncMongoengineObjectTypeOptions,
) = create_graphene_generic_class_async(ObjectType, ObjectTypeOptions)

(
    AsyncMongoengineInterfaceType,
    MongoengineInterfaceTypeOptions,
) = create_graphene_generic_class_async(Interface, InterfaceOptions)

AsyncGrapheneMongoengineObjectTypes = (
    AsyncMongoengineObjectType,
    AsyncMongoengineInterfaceType,
)



import pytest

from . import nodes, nodes_async
from .. import AsyncMongoengineConnectionField
from ..fields import MongoengineConnectionField


def test_article_field_args():
    field = MongoengineConnectionField(nodes.ArticleNode)

    field_args = {"id", "headline", "pub_date"}
    assert set(field.field_args.keys()) == field_args

    reference_args = {"editor", "reporter"}
    assert all(item in set(field.advance_args.keys()) for item in reference_args)

    default_args = {"after", "last", "first", "before"}
    args = field_args | reference_args | default_args
    assert set(field.args) == args


def test_reporter_field_args():
    field = MongoengineConnectionField(nodes.ReporterNode)

    field_args = {"id", "first_name", "last_name", "email", "awards"}
    assert set(field.field_args.keys()) == field_args


def test_editor_field_args():
    field = MongoengineConnectionField(nodes.EditorNode)

    field_args = {"id", "first_name", "last_name", "metadata", "seq"}
    assert set(field.field_args.keys()) == field_args


def test_field_args_with_property():
    field = MongoengineConnectionField(nodes.PublisherNode)

    field_args = ["id", "name"]
    assert set(field.field_args.keys()) == set(field_args)


def test_field_args_with_unconverted_field():
    field = MongoengineConnectionField(nodes.PublisherNode)

    field_args = ["id", "name"]
    assert set(field.field_args.keys()) == set(field_args)


@pytest.mark.asyncio
async def test_default_resolver_with_colliding_objects_field_async():
    field = AsyncMongoengineConnectionField(nodes_async.ErroneousModelAsyncNode)

    connection = await field.default_resolver(None, {})
    assert 0 == len(connection.iterable)


@pytest.mark.asyncio
async def test_default_resolver_connection_list_length_async(fixtures):
    field = AsyncMongoengineConnectionField(nodes_async.ArticleAsyncNode)

    connection = await field.default_resolver(None, {}, **{"first": 1})
    assert hasattr(connection, "list_length")
    assert connection.list_length == 1



from mongoengine import connect

from .models import Department, Employee, Role, Task

connect("graphene-mongo-example", host="mongomock://localhost", alias="default")


def init_db():
    # Create the fixtures
    engineering = Department(name="Engineering")
    engineering.save()

    hr = Department(name="Human Resources")
    hr.save()

    manager = Role(name="manager")
    manager.save()

    engineer = Role(name="engineer")
    engineer.save()

    debug = Task(name="Debug")
    test = Task(name="Test")

    tracy = Employee(name="Tracy", department=hr, roles=[engineer, manager], tasks=[])
    tracy.save()

    peter = Employee(
        name="Peter",
        department=engineering,
        leader=tracy,
        roles=[engineer],
        tasks=[debug, test],
    )
    peter.save()

    roy = Employee(
        name="Roy",
        department=engineering,
        leader=tracy,
        roles=[engineer],
        tasks=[debug],
    )
    roy.save()



import graphene
from graphene.relay import Node

from . import models
from . import types  # noqa: F401
from .models import ProfessorMetadata
from ..types import MongoengineObjectType


class PublisherNode(MongoengineObjectType):
    legal_name = graphene.String()
    bad_field = graphene.String()

    class Meta:
        model = models.Publisher
        only_fields = ("id", "name")
        interfaces = (Node,)


class ArticleNode(MongoengineObjectType):
    class Meta:
        model = models.Article
        interfaces = (Node,)


class EditorNode(MongoengineObjectType):
    class Meta:
        model = models.Editor
        interfaces = (Node,)


class EmbeddedArticleNode(MongoengineObjectType):
    class Meta:
        model = models.EmbeddedArticle
        interfaces = (Node,)


class PlayerNode(MongoengineObjectType):
    class Meta:
        model = models.Player
        interfaces = (Node,)
        filter_fields = {"first_name": ["istartswith", "in"]}


class ReporterNode(MongoengineObjectType):
    class Meta:
        model = models.Reporter
        interfaces = (Node,)


class ParentNode(MongoengineObjectType):
    class Meta:
        model = models.Parent
        interfaces = (Node,)


class ChildNode(MongoengineObjectType):
    class Meta:
        model = models.Child
        interfaces = (Node,)


class ChildRegisteredBeforeNode(MongoengineObjectType):
    class Meta:
        model = models.ChildRegisteredBefore
        interfaces = (Node,)


class ChildRegisteredAfterNode(MongoengineObjectType):
    class Meta:
        model = models.ChildRegisteredAfter
        interfaces = (Node,)


class ParentWithRelationshipNode(MongoengineObjectType):
    class Meta:
        model = models.ParentWithRelationship
        interfaces = (Node,)


class ProfessorMetadataNode(MongoengineObjectType):
    class Meta:
        model = ProfessorMetadata
        interfaces = (graphene.Node,)


class ProfessorVectorNode(MongoengineObjectType):
    class Meta:
        model = models.ProfessorVector
        interfaces = (Node,)


class ErroneousModelNode(MongoengineObjectType):
    class Meta:
        model = models.ErroneousModel
        interfaces = (Node,)


class BarNode(MongoengineObjectType):
    class Meta:
        model = models.Bar
        interfaces = (Node,)


class FooNode(MongoengineObjectType):
    class Meta:
        model = models.Foo
        interfaces = (Node,)



from graphene import Enum

from graphene_mongo.utils import ExecutorEnum


class Registry(object):
    def __init__(self):
        self._registry = {}
        self._registry_async = {}
        self._registry_string_map = {}
        self._registry_async_string_map = {}
        self._registry_enum = {}

    def register(self, cls):
        from .types import GrapheneMongoengineObjectTypes
        from .types_async import AsyncGrapheneMongoengineObjectTypes

        assert (
            issubclass(cls, GrapheneMongoengineObjectTypes)
            or issubclass(cls, AsyncGrapheneMongoengineObjectTypes)
        ), 'Only Mongoengine/Async Mongoengine object types can be registered, received "{}"'.format(
            cls.__name__
        )
        assert cls._meta.registry == self, "Registry for a Model have to match."
        if issubclass(cls, GrapheneMongoengineObjectTypes):
            self._registry[cls._meta.model] = cls
            self._registry_string_map[cls.__name__] = cls._meta.model.__name__
        else:
            self._registry_async[cls._meta.model] = cls
            self._registry_async_string_map[cls.__name__] = cls._meta.model.__name__

        # Rescan all fields
        for model, cls in self._registry.items():
            cls.rescan_fields()

    def register_enum(self, cls):
        from enum import EnumMeta

        assert isinstance(
            cls, EnumMeta
        ), f'Only EnumMeta can be registered, received "{cls.__name__}"'
        if not cls.__name__.endswith("Enum"):
            name = cls.__name__ + "Enum"
        else:
            name = cls.__name__
        cls.__name__ = name
        self._registry_enum[cls] = Enum.from_enum(cls)

    def get_type_for_model(self, model, executor: ExecutorEnum = ExecutorEnum.SYNC):
        if executor == ExecutorEnum.SYNC:
            return self._registry.get(model)
        else:
            return self._registry_async.get(model)

    def check_enum_already_exist(self, cls):
        return cls in self._registry_enum

    def get_type_for_enum(self, cls):
        return self._registry_enum.get(cls)


registry = None
async_registry = None
inputs_registry = None
async_inputs_registry = None


def get_inputs_registry():
    global inputs_registry
    if not inputs_registry:
        inputs_registry = Registry()
    return inputs_registry


def get_inputs_async_registry():
    global async_inputs_registry
    if not async_inputs_registry:
        async_inputs_registry = Registry()
    return async_inputs_registry


def get_global_registry():
    global registry
    if not registry:
        registry = Registry()
    return registry


def get_global_async_registry():
    global async_registry
    if not async_registry:
        async_registry = Registry()
    return async_registry


def reset_global_registry():
    global registry
    global inputs_registry
    registry = None
    inputs_registry = None


def reset_global_async_registry():
    global async_registry
    global async_inputs_registry
    async_registry = None
    async_inputs_registry = None



from .fields import MongoengineConnectionField
from .fields_async import AsyncMongoengineConnectionField

from .types import MongoengineObjectType, MongoengineInputType, MongoengineInterfaceType
from .types_async import AsyncMongoengineObjectType

__version__ = "0.1.1"

__all__ = [
    "__version__",
    "MongoengineObjectType",
    "AsyncMongoengineObjectType",
    "MongoengineInputType",
    "MongoengineInterfaceType",
    "MongoengineConnectionField",
    "AsyncMongoengineConnectionField",
]



from graphene import relay

from graphene_mongo import MongoengineObjectType
from .models import Bookmark, Category


class CategoryType(MongoengineObjectType):
    class Meta:
        model = Category
        interfaces = (relay.Node,)


class BookmarkType(MongoengineObjectType):
    class Meta:
        model = Bookmark
        interfaces = (relay.Node,)



import pytest
from .models import Bike, Shop


def fixture_bike_data():
    Bike.drop_collection()
    bike_one = Bike(
        name="Level R",
        brand="Mondraker",
        year="2020",
        size=["S", "M", "L", "XL"],
        wheel_size=27.5,
        type="MTB",
    )
    bike_one.save()

    bike_two = Bike(
        name="CAADX ULTEGRA",
        brand="Cannondale",
        year="2019",
        size=["46", "51", "54", "58"],
        wheel_size=28,
        type="Gravel",
    )
    bike_two.save()

    bike_three = Bike(
        id="507f1f77bcf86cd799439011",
        name="Moterra Neo",
        brand="Cannondale",
        year="2019",
        size=["M", "L", "XL"],
        wheel_size=29,
        type="EBike",
    )
    bike_three.save()


def fixture_shop_data():
    Shop.drop_collection()
    shop_one = Shop(
        name="Big Wheel Bicycles",
        address="2438 Hart Ridge Road",
        website="https://www.bigwheelbike.test",
    )
    shop_one.save()
    shop_two = Shop(
        name="Bike Tech",
        address="2175 Pearl Street",
        website="https://www.biketech.test",
    )
    shop_two.save()


@pytest.fixture(scope="module")
def fixtures_data():
    fixture_bike_data()
    fixture_shop_data()

    return True



from datetime import datetime
from mongoengine import Document, EmbeddedDocument
from mongoengine.fields import (
    DateTimeField,
    EmbeddedDocumentField,
    ListField,
    ReferenceField,
    StringField,
)


class Department(Document):
    meta = {"collection": "department"}
    name = StringField()


class Role(Document):
    meta = {"collection": "role"}
    name = StringField()


class Task(EmbeddedDocument):
    name = StringField()
    deadline = DateTimeField(default=datetime.now)


class Employee(Document):
    meta = {"collection": "employee"}
    name = StringField()
    hired_on = DateTimeField(default=datetime.now)
    department = ReferenceField(Department)
    roles = ListField(ReferenceField(Role))
    leader = ReferenceField("Employee")
    tasks = ListField(EmbeddedDocumentField(Task))



import graphene
import mongoengine
from pytest import raises

from .models import (
    Article,
    Editor,
    EmbeddedArticle,
    EmbeddedFoo,
    Player,
    Reporter,
    ProfessorMetadata,
    ProfessorVector,
    Publisher,
)
from .. import registry
from .. import advanced_types
from ..converter import convert_mongoengine_field
from ..fields import MongoengineConnectionField
from ..types import MongoengineObjectType


def assert_conversion(mongoengine_field, graphene_field, *args, **kwargs):
    field = mongoengine_field(*args, **kwargs)
    graphene_type = convert_mongoengine_field(field)
    assert isinstance(graphene_type, graphene_field)
    field = graphene_type.Field()
    return field


def test_should_unknown_mongoengine_field_raise_exception():
    with raises(Exception) as excinfo:
        convert_mongoengine_field(None)
    assert "Don't know how to convert the MongoEngine field" in str(excinfo)


def test_should_email_convert_string():
    assert_conversion(mongoengine.EmailField, graphene.String)


def test_should_string_convert_string():
    assert_conversion(mongoengine.StringField, graphene.String)


def test_should_url_convert_string():
    assert_conversion(mongoengine.URLField, graphene.String)


def test_should_uuid_convert_id():
    assert_conversion(mongoengine.UUIDField, graphene.ID)


def test_sould_int_convert_int():
    assert_conversion(mongoengine.IntField, graphene.Int)


def test_sould_long_convert_int():
    assert_conversion(mongoengine.LongField, graphene.Int)


def test_sould_sequence_convert_field():
    assert_conversion(mongoengine.SequenceField, graphene.Int)


def test_should_object_id_convert_id():
    assert_conversion(mongoengine.ObjectIdField, graphene.ID)


def test_should_boolean_convert_boolean():
    assert_conversion(mongoengine.BooleanField, graphene.Boolean)


def test_should_decimal_convert_decimal():
    assert_conversion(mongoengine.DecimalField, graphene.Decimal)


def test_should_float_convert_float():
    assert_conversion(mongoengine.FloatField, graphene.Float)


def test_should_decimal128_convert_decimal():
    assert_conversion(mongoengine.Decimal128Field, graphene.Decimal)


def test_should_datetime_convert_datetime():
    assert_conversion(mongoengine.DateTimeField, graphene.DateTime)


def test_should_dict_convert_json():
    assert_conversion(mongoengine.DictField, graphene.JSONString)


def test_should_map_convert_json():
    assert_conversion(mongoengine.MapField, graphene.JSONString, field=mongoengine.StringField())


def test_should_point_convert_field():
    graphene_type = convert_mongoengine_field(mongoengine.PointField())
    assert isinstance(graphene_type, graphene.Field)
    assert graphene_type.type == advanced_types.PointFieldType
    assert isinstance(graphene_type.type.type, graphene.String)
    assert isinstance(graphene_type.type.coordinates, graphene.List)


def test_should_polygon_covert_field():
    graphene_type = convert_mongoengine_field(mongoengine.PolygonField())
    assert isinstance(graphene_type, graphene.Field)
    assert graphene_type.type == advanced_types.PolygonFieldType
    assert isinstance(graphene_type.type.type, graphene.String)
    assert isinstance(graphene_type.type.coordinates, graphene.List)


def test_should_multipolygon_convert_field():
    graphene_type = convert_mongoengine_field(mongoengine.MultiPolygonField())
    assert isinstance(graphene_type, graphene.Field)
    assert graphene_type.type == advanced_types.MultiPolygonFieldType
    assert isinstance(graphene_type.type.type, graphene.String)
    assert isinstance(graphene_type.type.coordinates, graphene.List)


def test_should_file_convert_field():
    graphene_type = convert_mongoengine_field(mongoengine.FileField())
    assert isinstance(graphene_type, graphene.Field)
    assert graphene_type.type == advanced_types.FileFieldType


def test_should_field_convert_list():
    assert_conversion(mongoengine.ListField, graphene.List, field=mongoengine.StringField())


def test_should_geo_convert_list():
    assert_conversion(mongoengine.GeoPointField, graphene.List, field=mongoengine.FloatField())


def test_should_reference_convert_dynamic():
    class E(MongoengineObjectType):
        class Meta:
            model = Editor
            interfaces = (graphene.Node,)

    dynamic_field = convert_mongoengine_field(EmbeddedArticle._fields["editor"], E._meta.registry)
    assert isinstance(dynamic_field, graphene.Dynamic)
    graphene_type = dynamic_field.get_type()
    assert isinstance(graphene_type, graphene.Field)
    assert graphene_type.type == E


def test_should_lazy_reference_convert_dynamic():
    class P(MongoengineObjectType):
        class Meta:
            model = Publisher
            interfaces = (graphene.Node,)

    dynamic_field = convert_mongoengine_field(Editor._fields["company"], P._meta.registry)

    assert isinstance(dynamic_field, graphene.Dynamic)
    graphene_type = dynamic_field.get_type()
    assert isinstance(graphene_type, graphene.Field)
    assert graphene_type.type == P


def test_should_embedded_convert_dynamic():
    class PM(MongoengineObjectType):
        class Meta:
            model = ProfessorMetadata
            interfaces = (graphene.Node,)

    dynamic_field = convert_mongoengine_field(
        ProfessorVector._fields["metadata"], PM._meta.registry
    )
    assert isinstance(dynamic_field, graphene.Dynamic)
    graphene_type = dynamic_field.get_type()
    assert isinstance(graphene_type, graphene.Field)
    assert graphene_type.type == PM


def test_should_convert_none():
    registry.reset_global_registry()
    dynamic_field = convert_mongoengine_field(
        EmbeddedArticle._fields["editor"], registry.get_global_registry()
    )
    assert isinstance(dynamic_field, graphene.Dynamic)
    graphene_type = dynamic_field.get_type()
    assert graphene_type is None


def test_should_convert_none_lazily():
    registry.reset_global_registry()
    dynamic_field = convert_mongoengine_field(
        Editor._fields["company"], registry.get_global_registry()
    )
    assert isinstance(dynamic_field, graphene.Dynamic)
    graphene_type = dynamic_field.get_type()
    assert graphene_type is None


def test_should_list_of_reference_convert_list():
    class A(MongoengineObjectType):
        class Meta:
            model = Article

    graphene_field = convert_mongoengine_field(Reporter._fields["articles"], A._meta.registry)
    assert isinstance(graphene_field, graphene.List)
    dynamic_field = graphene_field.get_type()
    assert dynamic_field._of_type == A


def test_should_list_of_generic_reference_covert_list():
    class A(MongoengineObjectType):
        class Meta:
            model = Article

    class E(MongoengineObjectType):
        class Meta:
            model = Editor

    class R(MongoengineObjectType):
        class Meta:
            model = Reporter

    generic_references_field = convert_mongoengine_field(
        Reporter._fields["generic_references"], registry.get_global_registry()
    )
    assert isinstance(generic_references_field, graphene.List)
    field = generic_references_field.get_type()
    assert field._of_type._meta.types == (A, E)


def test_should_list_of_embedded_convert_list():
    class E(MongoengineObjectType):
        class Meta:
            model = EmbeddedArticle

    graphene_field = convert_mongoengine_field(
        Reporter._fields["embedded_articles"], E._meta.registry
    )
    assert isinstance(graphene_field, graphene.List)
    dynamic_field = graphene_field.get_type()
    assert dynamic_field._of_type == E


def test_should_embedded_list_convert_list():
    class E(MongoengineObjectType):
        class Meta:
            model = EmbeddedArticle

    graphene_field = convert_mongoengine_field(
        Reporter._fields["embedded_list_articles"], E._meta.registry
    )
    assert isinstance(graphene_field, graphene.List)
    dynamic_field = graphene_field.get_type()
    assert dynamic_field._of_type == E


def test_should_self_reference_convert_dynamic():
    class P(MongoengineObjectType):
        class Meta:
            model = Player
            interfaces = (graphene.Node,)

    dynamic_field = convert_mongoengine_field(Player._fields["opponent"], P._meta.registry)
    assert isinstance(dynamic_field, graphene.Dynamic)
    graphene_type = dynamic_field.get_type()
    assert isinstance(graphene_type, graphene.Field)
    assert graphene_type.type == P

    graphene_field = convert_mongoengine_field(Player._fields["players"], P._meta.registry)
    assert isinstance(graphene_field, MongoengineConnectionField)


def test_should_list_of_self_reference_convert_list():
    class A(MongoengineObjectType):
        class Meta:
            model = Article

    class P(MongoengineObjectType):
        class Meta:
            model = Player

    graphene_field = convert_mongoengine_field(Player._fields["players"], P._meta.registry)
    assert isinstance(graphene_field, graphene.List)
    dynamic_field = graphene_field.get_type()
    assert dynamic_field._of_type == P


def test_should_description_convert_common_metadata():
    class A(MongoengineObjectType):
        class Meta:
            model = Article

    headline_field = convert_mongoengine_field(Article._fields["headline"], A._meta.registry)
    assert headline_field.kwargs["description"] == "The article headline."

    pubDate_field = convert_mongoengine_field(Article._fields["pub_date"], A._meta.registry)
    assert pubDate_field.kwargs["description"] == "Publication Date\nThe date of first press."

    firstName_field = convert_mongoengine_field(Editor._fields["first_name"], A._meta.registry)
    assert firstName_field.kwargs["description"] == "Editor's first name.\n(fname)"

    metadata_field = convert_mongoengine_field(Editor._fields["metadata"], A._meta.registry)
    assert metadata_field.kwargs["description"] == "Arbitrary metadata."


def test_should_description_convert_reference_metadata():
    class A(MongoengineObjectType):
        class Meta:
            model = Article

    class E(MongoengineObjectType):
        class Meta:
            model = Editor

    editor_field = convert_mongoengine_field(Article._fields["editor"], A._meta.registry).get_type()
    assert editor_field.description == "An Editor of a publication."


def test_should_generic_reference_convert_union():
    class A(MongoengineObjectType):
        class Meta:
            model = Article

    class E(MongoengineObjectType):
        class Meta:
            model = Editor

    class R(MongoengineObjectType):
        class Meta:
            model = Reporter

    generic_reference_field = convert_mongoengine_field(
        Reporter._fields["generic_reference"], registry.get_global_registry()
    )
    assert isinstance(generic_reference_field, graphene.Field)
    if not Reporter._fields["generic_reference"].required:
        assert isinstance(generic_reference_field.type(), graphene.Union)
        assert generic_reference_field.type()._meta.types == (A, E)
    else:
        assert issubclass(generic_reference_field.type.of_type, graphene.Union)
        assert generic_reference_field.type.of_type._meta.types == (A, E)


def test_should_generic_embedded_document_convert_union():
    class D(MongoengineObjectType):
        class Meta:
            model = EmbeddedArticle

    class F(MongoengineObjectType):
        class Meta:
            model = EmbeddedFoo

    class A(MongoengineObjectType):
        class Meta:
            model = Article

    class E(MongoengineObjectType):
        class Meta:
            model = Editor

    class R(MongoengineObjectType):
        class Meta:
            model = Reporter

    generic_embedded_document = convert_mongoengine_field(
        Reporter._fields["generic_embedded_document"], registry.get_global_registry()
    )
    assert isinstance(generic_embedded_document, graphene.Field)
    assert isinstance(generic_embedded_document.type(), graphene.Union)
    assert generic_embedded_document.type()._meta.types == (D, F)



import mongoengine
from graphene.test import Client

from examples.falcon_mongoengine.schema import schema
from .fixtures import fixtures_data

mongoengine.connect("graphene-mongo-test", host="mongomock://localhost", alias="default")


def test_category_last_1_item_query(fixtures_data):
    query = """
               {
               categories(last: 1){
                   edges {
                   node {
                       name
                       color 
                       }
                   }
               }
           }"""

    expected = {
        "data": {
            "categories": {
                "edges": [
                    {
                        "node": {
                            "name": "Work",
                            "color": "#1769ff",
                        }
                    }
                ]
            }
        }
    }

    client = Client(schema)
    result = client.execute(query)
    assert result == expected


def test_category_filter_item_query(fixtures_data):
    query = """
               {
               categories(name: "Work"){
                   edges {
                   node {
                       name
                       color 
                       }
                   }
               }
           }"""

    expected = {"data": {"categories": {"edges": [{"node": {"name": "Work", "color": "#1769ff"}}]}}}

    client = Client(schema)
    result = client.execute(query)
    assert result == expected


def test_bookmarks_first_2_items_query(fixtures_data):
    query = """
               {
               bookmarks(first: 2){
                   edges {
                   node {
                       name
                       url
                       category {
                            name
                            color
                            }
                       tags
                       }
                   }
               }
           }"""

    expected = {
        "data": {
            "bookmarks": {
                "edges": [
                    {
                        "node": {
                            "name": "Travel tips",
                            "url": "https://www.traveltips.test",
                            "category": {"name": "Travel", "color": "#ed008c"},
                            "tags": ["travel", "tips", "howto"],
                        }
                    },
                    {
                        "node": {
                            "name": "DIY vacation",
                            "url": "https://www.diyvacation.test",
                            "category": {"name": "Travel", "color": "#ed008c"},
                            "tags": ["travel", "diy", "holiday", "vacation"],
                        }
                    },
                ]
            }
        }
    }

    client = Client(schema)
    result = client.execute(query)
    assert result == expected


def test_bookmarks_filter_items_query(fixtures_data):
    query = """
               {
               bookmarks(first: 1, name: "Awesome python"){
                   edges {
                   node {
                       name
                       url
                       category {
                            name
                            color
                            }
                       tags
                       }
                   }
               }
           }"""

    expected = {
        "data": {
            "bookmarks": {
                "edges": [
                    {
                        "node": {
                            "name": "Awesome python",
                            "url": "https://awesomelists.top/#repos/vinta/awesome-python",
                            "category": {"name": "Work", "color": "#1769ff"},
                            "tags": ["python", "dev", "awesome", "tutorial"],
                        }
                    }
                ]
            }
        }
    }

    client = Client(schema)
    result = client.execute(query)
    assert result == expected



import graphene
from graphene.relay import Node

from graphene_mongo import MongoengineConnectionField, MongoengineObjectType
from .models import Department as DepartmentModel
from .models import Employee as EmployeeModel
from .models import Role as RoleModel
from .models import Task as TaskModel


class Department(MongoengineObjectType):
    class Meta:
        model = DepartmentModel
        interfaces = (Node,)


class Role(MongoengineObjectType):
    class Meta:
        model = RoleModel
        interfaces = (Node,)
        filter_fields = {
            "name": [
                "exact",
                "icontains",
                "istartswith",
            ]
        }


class Task(MongoengineObjectType):
    class Meta:
        model = TaskModel
        interfaces = (Node,)


class Employee(MongoengineObjectType):
    class Meta:
        model = EmployeeModel
        interfaces = (Node,)
        filter_fields = {
            "name": [
                "exact",
                "icontains",
                "istartswith",
            ]
        }


class Query(graphene.ObjectType):
    node = Node.Field()
    all_employees = MongoengineConnectionField(Employee)
    all_roles = MongoengineConnectionField(Role)
    role = graphene.Field(Role)


schema = graphene.Schema(query=Query, types=[Department, Employee, Role])



import pytest
from django.urls import reverse
from django.test import RequestFactory
from graphene.test import Client
from .schema import schema
from .fixtures import fixtures_data


def test_bikes_first_item_query(fixtures_data):
    query = """
               {
               bikes(first: 1){
                   edges {
                   node {
                       name
                       brand
                       year
                       size
                       wheelSize
                       type
                       }
                   }
               }
           }"""

    expected = {
        "data": {
            "bikes": {
                "edges": [
                    {
                        "node": {
                            "name": "Level R",
                            "brand": "Mondraker",
                            "year": "2020",
                            "size": ["S", "M", "L", "XL"],
                            "wheelSize": 27.5,
                            "type": "MTB",
                        }
                    }
                ]
            }
        }
    }

    client = Client(schema)
    result = client.execute(query)
    assert result == expected


def test_bikes_filter_by_type_item_query(fixtures_data):
    query = """
               {
               bikes(first: 2, type: "Gravel"){
                   edges {
                   node {
                       name
                       brand
                       year
                       size
                       wheelSize
                       type
                       }
                   }
               }
           }"""

    expected = {
        "data": {
            "bikes": {
                "edges": [
                    {
                        "node": {
                            "name": "CAADX ULTEGRA",
                            "brand": "Cannondale",
                            "year": "2019",
                            "size": ["46", "51", "54", "58"],
                            "wheelSize": 28,
                            "type": "Gravel",
                        }
                    }
                ]
            }
        }
    }

    client = Client(schema)
    result = client.execute(query)
    assert result == expected


def test_shop_data_query(fixtures_data):
    query = """{
               shopList{
                    name
                    address
                    website
                    }
            }"""

    expected = {
        "data": {
            "shopList": [
                {
                    "name": "Big Wheel Bicycles",
                    "address": "2438 Hart Ridge Road",
                    "website": "https://www.bigwheelbike.test",
                },
                {
                    "name": "Bike Tech",
                    "address": "2175 Pearl Street",
                    "website": "https://www.biketech.test",
                },
            ]
        }
    }

    client = Client(schema)
    result = client.execute(query)
    assert result == expected


@pytest.mark.django_db
def test_create_bike_mutation():
    query = """
                    mutation {
                        createBike(bikeData:{
                                name:"Bullhorn",
                                brand:"Pegas",
                                year: "2019",
                                size: ["56", "58" ],
                                wheelSize: 28,
                                type: "Fixie"
                                }) {
                            bike {
                                name
                                brand
                                year
                                size
                                wheelSize
                                type
                            }
                        }
                    }
                  """

    expected = {
        "data": {
            "createBike": {
                "bike": {
                    "name": "Bullhorn",
                    "brand": "Pegas",
                    "year": "2019",
                    "size": ["56", "58"],
                    "wheelSize": 28,
                    "type": "Fixie",
                }
            }
        }
    }

    factory = RequestFactory()
    request = factory.post(reverse("graphql-query"))
    client = Client(schema)
    result = client.execute(query, context=request)
    assert result == expected


@pytest.mark.django_db
def test_update_bike_mutation():
    query = """
                    mutation {
                        updateBike(bikeData:{
                                id: "507f1f77bcf86cd799439011",
                                name:"Moterra Neo Updated",
                                year: "2020",
                                wheelSize: 27.5,
                                type: "EBike Updated"
                                }) {
                            bike {
                                name
                                brand
                                year
                                size
                                wheelSize
                                type
                            }
                        }
                    }
                  """

    expected = {
        "data": {
            "updateBike": {
                "bike": {
                    "name": "Moterra Neo Updated",
                    "brand": "Cannondale",
                    "year": "2020",
                    "size": ["M", "L", "XL"],
                    "wheelSize": 27.5,
                    "type": "EBike Updated",
                }
            }
        }
    }

    factory = RequestFactory()
    request = factory.post(reverse("graphql-query"))
    client = Client(schema)
    result = client.execute(query, context=request)
    print(result)
    assert result == expected


@pytest.mark.django_db
def test_delete_bike_mutation():
    query = """
                mutation {
                    deleteBike(id: "507f1f77bcf86cd799439011") {
                           success
                    }
                }
              """

    expected = {"data": {"deleteBike": {"success": True}}}

    factory = RequestFactory()
    request = factory.post(reverse("graphql-query"))
    client = Client(schema)
    result = client.execute(query, context=request)
    assert result == expected



from django.urls import path
from django.views.decorators.csrf import csrf_exempt
from graphene_django.views import GraphQLView

urlpatterns = [
    path("graphql", csrf_exempt(GraphQLView.as_view(graphiql=True)), name="graphql-query")
]



from collections import OrderedDict

import graphene
import mongoengine
from graphene.relay import Connection, Node
from graphene.types.inputobjecttype import InputObjectType, InputObjectTypeOptions
from graphene.types.interface import Interface, InterfaceOptions
from graphene.types.objecttype import ObjectType, ObjectTypeOptions
from graphene.types.utils import yank_fields_from_attrs
from graphene.utils.str_converters import to_snake_case

from graphene_mongo import MongoengineConnectionField
from .converter import convert_mongoengine_field
from .registry import Registry, get_global_registry, get_inputs_registry
from .utils import (
    ExecutorEnum,
    get_model_fields,
    get_query_fields,
    is_valid_mongoengine_model,
    sync_to_async,
)


def construct_fields(
    model,
    registry,
    only_fields,
    exclude_fields,
    non_required_fields,
    executor: ExecutorEnum = ExecutorEnum.SYNC,
):
    """
    Args:
        model (mongoengine.Document):
        registry (.registry.Registry):
        only_fields ([str]):
        exclude_fields ([str]):
        executor : ExecutorEnum

    Returns:
        (OrderedDict, OrderedDict): converted fields and self reference fields.

    """
    _model_fields = get_model_fields(model)
    fields = OrderedDict()
    self_referenced = OrderedDict()
    for name, field in _model_fields.items():
        is_not_in_only = only_fields and name not in only_fields
        is_excluded = name in exclude_fields
        if is_not_in_only or is_excluded:
            # We skip this field if we specify required_fields and is not
            # in there. Or when we exclude this field in exclude_fields
            continue
        if isinstance(field, mongoengine.ListField):
            if not field.field:
                continue
            # Take care of list of self-reference.
            document_type_obj = field.field.__dict__.get("document_type_obj", None)
            if (
                document_type_obj == model._class_name
                or isinstance(document_type_obj, model)
                or document_type_obj == model
            ):
                self_referenced[name] = field
                continue
        converted = convert_mongoengine_field(field, registry, executor)
        if not converted:
            continue
        else:
            if name in non_required_fields and "required" in converted.kwargs:
                converted.kwargs["required"] = False
        fields[name] = converted

    return fields, self_referenced


def construct_self_referenced_fields(self_referenced, registry, executor=ExecutorEnum.SYNC):
    fields = OrderedDict()
    for name, field in self_referenced.items():
        converted = convert_mongoengine_field(field, registry, executor)
        if not converted:
            continue
        fields[name] = converted

    return fields


def create_graphene_generic_class(object_type, option_type):
    class MongoengineGenericObjectTypeOptions(option_type):
        model = None
        registry = None  # type: Registry
        connection = None
        filter_fields = ()
        non_required_fields = ()
        order_by = None

    class GrapheneMongoengineGenericType(object_type):
        @classmethod
        def __init_subclass_with_meta__(
            cls,
            model=None,
            registry=None,
            skip_registry=False,
            only_fields=(),
            required_fields=(),
            exclude_fields=(),
            non_required_fields=(),
            filter_fields=None,
            non_filter_fields=(),
            connection=None,
            connection_class=None,
            use_connection=None,
            connection_field_class=None,
            interfaces=(),
            _meta=None,
            order_by=None,
            **options,
        ):
            assert is_valid_mongoengine_model(model), (
                "The attribute model in {}.Meta must be a valid Mongoengine Model. "
                'Received "{}" instead.'
            ).format(cls.__name__, type(model))

            if not registry:
                # input objects shall be registred in a separated registry
                if issubclass(cls, InputObjectType):
                    registry = get_inputs_registry()
                else:
                    registry = get_global_registry()

            assert isinstance(registry, Registry), (
                "The attribute registry in {}.Meta needs to be an instance of "
                'Registry({}), received "{}".'
            ).format(object_type, cls.__name__, registry)
            converted_fields, self_referenced = construct_fields(
                model, registry, only_fields, exclude_fields, non_required_fields
            )
            mongoengine_fields = yank_fields_from_attrs(converted_fields, _as=graphene.Field)
            if use_connection is None and interfaces:
                use_connection = any((issubclass(interface, Node) for interface in interfaces))

            if use_connection and not connection:
                # We create the connection automatically
                if not connection_class:
                    connection_class = Connection

                connection = connection_class.create_type(
                    "{}Connection".format(options.get("name") or cls.__name__), node=cls
                )

            if connection is not None:
                assert issubclass(connection, Connection), (
                    "The attribute connection in {}.Meta must be of type Connection. "
                    'Received "{}" instead.'
                ).format(cls.__name__, type(connection))

            if connection_field_class is not None:
                assert issubclass(connection_field_class, graphene.ConnectionField), (
                    "The attribute connection_field_class in {}.Meta must be of type graphene.ConnectionField. "
                    'Received "{}" instead.'
                ).format(cls.__name__, type(connection_field_class))
            else:
                connection_field_class = MongoengineConnectionField

            if _meta:
                assert isinstance(_meta, MongoengineGenericObjectTypeOptions), (
                    "_meta must be an instance of MongoengineGenericObjectTypeOptions, "
                    "received {}"
                ).format(_meta.__class__)
            else:
                _meta = MongoengineGenericObjectTypeOptions(option_type)

            _meta.model = model
            _meta.registry = registry
            _meta.fields = mongoengine_fields
            _meta.filter_fields = filter_fields
            _meta.non_filter_fields = non_filter_fields
            _meta.connection = connection
            _meta.connection_field_class = connection_field_class
            # Save them for later
            _meta.only_fields = only_fields
            _meta.required_fields = required_fields
            _meta.exclude_fields = exclude_fields
            _meta.non_required_fields = non_required_fields
            _meta.order_by = order_by

            super(GrapheneMongoengineGenericType, cls).__init_subclass_with_meta__(
                _meta=_meta, interfaces=interfaces, **options
            )

            if not skip_registry:
                registry.register(cls)
                # Notes: Take care list of self-reference fields.
                converted_fields = construct_self_referenced_fields(self_referenced, registry)
                if converted_fields:
                    mongoengine_fields = yank_fields_from_attrs(
                        converted_fields, _as=graphene.Field
                    )
                    cls._meta.fields.update(mongoengine_fields)
                    registry.register(cls)

        @classmethod
        def rescan_fields(cls):
            """Attempts to rescan fields and will insert any not converted initially"""

            converted_fields, self_referenced = construct_fields(
                cls._meta.model,
                cls._meta.registry,
                cls._meta.only_fields,
                cls._meta.exclude_fields,
                cls._meta.non_required_fields,
            )

            mongoengine_fields = yank_fields_from_attrs(converted_fields, _as=graphene.Field)

            # The initial scan should take precedence
            for field in mongoengine_fields:
                if field not in cls._meta.fields:
                    cls._meta.fields.update({field: mongoengine_fields[field]})
            # Self-referenced fields can't change between scans!

        @classmethod
        def is_type_of(cls, root, info):
            if isinstance(root, cls):
                return True
            # XXX: Take care FileField
            if isinstance(root, mongoengine.GridFSProxy):
                return True
            if not is_valid_mongoengine_model(type(root)):
                raise Exception(('Received incompatible instance "{}".').format(root))
            return isinstance(root, cls._meta.model)

        @classmethod
        async def get_node(cls, info, id):
            required_fields = list()
            for field in cls._meta.required_fields:
                if field in cls._meta.model._fields_ordered:
                    required_fields.append(field)
            queried_fields = get_query_fields(info)
            if cls._meta.name in queried_fields:
                queried_fields = queried_fields[cls._meta.name]
            for field in queried_fields:
                if to_snake_case(field) in cls._meta.model._fields_ordered:
                    required_fields.append(to_snake_case(field))
            required_fields = list(set(required_fields))
            return await sync_to_async(
                cls._meta.model.objects.no_dereference().only(*required_fields).get,
            )(pk=id)

        def resolve_id(self, info):
            return str(self.id)

    return GrapheneMongoengineGenericType, MongoengineGenericObjectTypeOptions


MongoengineObjectType, MongoengineObjectTypeOptions = create_graphene_generic_class(
    ObjectType, ObjectTypeOptions
)
MongoengineInterfaceType, MongoengineInterfaceTypeOptions = create_graphene_generic_class(
    Interface, InterfaceOptions
)
MongoengineInputType, MongoengineInputTypeOptions = create_graphene_generic_class(
    InputObjectType, InputObjectTypeOptions
)

GrapheneMongoengineObjectTypes = (
    MongoengineObjectType,
    MongoengineInputType,
    MongoengineInterfaceType,
)



import falcon
from mongoengine import connect
from .api import GraphQLResource, HelloWorldResource

connect("bookmarks_db", host="127.0.0.1", port=27017)
app = application = falcon.API()

helloWorld = HelloWorldResource()
graphQL = GraphQLResource()

app.add_route("/", helloWorld)
app.add_route("/graphql", graphQL)



from mongoengine import Document, CASCADE
from mongoengine.fields import StringField, ListField, ReferenceField


class Category(Document):
    meta = {"collection": "category"}
    name = StringField(max_length=140, required=True)
    color = StringField(max_length=7, required=True)


class Bookmark(Document):
    meta = {"collection": "bookmark"}
    name = StringField(required=True)
    url = StringField(required=True)
    category = ReferenceField("Category", reverse_delete_rule=CASCADE)
    tags = ListField(StringField(max_length=50))



import graphene

from . import types
from .models import Article, Child, Reporter
from ..utils import get_model_fields, get_query_fields, is_valid_mongoengine_model


def test_get_model_fields_no_duplication():
    reporter_fields = get_model_fields(Reporter)
    reporter_name_set = set(reporter_fields)
    assert len(reporter_fields) == len(reporter_name_set)


def test_get_model_fields_excluding():
    reporter_fields = get_model_fields(Reporter, excluding=["first_name", "last_name"])
    reporter_name_set = set(reporter_fields)
    assert all(
        field in reporter_name_set
        for field in [
            "id",
            "email",
            "articles",
            "embedded_articles",
            "embedded_list_articles",
            "awards",
        ]
    )


def test_get_model_relation_fields():
    article_fields = get_model_fields(Article)
    assert all(field in set(article_fields) for field in ["editor", "reporter"])


def test_get_base_model_fields():
    child_fields = get_model_fields(Child)
    assert all(field in set(child_fields) for field in ["bar", "baz"])


def test_is_valid_mongoengine_mode():
    assert is_valid_mongoengine_model(Reporter)


def test_get_query_fields():
    # Grab ResolveInfo objects from resolvers and set as nonlocal variables outside
    # Can't assert within resolvers, as the resolvers may not be run if there is an exception
    class Query(graphene.ObjectType):
        child = graphene.Field(types.ChildType)
        children = graphene.List(types.ChildUnionType)

        def resolve_child(self, info, *args, **kwargs):
            test_get_query_fields.child_info = info

        def resolve_children(self, info, *args, **kwargs):
            test_get_query_fields.children_info = info

    query = """
        query Query {
            child {
                bar
                ...testFragment
            }
            children {
                ... on ChildType{
                    baz
                    ...testFragment
                }
                ... on AnotherChildType {
                    qux
                }
            }
        }

        fragment testFragment on ChildType {
            loc {
                type
                coordinates
            }
        }
    """

    schema = graphene.Schema(query=Query)
    schema.execute(query)

    assert get_query_fields(test_get_query_fields.child_info) == {
        "bar": {},
        "loc": {
            "type": {},
            "coordinates": {},
        },
    }

    assert get_query_fields(test_get_query_fields.children_info) == {
        "ChildType": {
            "baz": {},
            "loc": {
                "type": {},
                "coordinates": {},
            },
        },
        "AnotherChildType": {
            "qux": {},
        },
    }
